/*
 * Indie Game Jame 2 - Ken's "Broken Dodgers"
 * 2004/3/15 checker@d6.com
 * 2004/3/16 ken@demarest.com
 */
 
/*

Other thoughts:
- defend a critter as it crosses the screen
- walk in the shadow of larger, slow moving things
- touch power ups to get health, 
- some things are not harmfull, can be pushed to then move other dangerous stuff
	- some (or all) of these are fixed
- generators shooting out evil
	- lines
	- clouds
	- duration on life of evil
	- basically, any particle generator
*/ 
 
 
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_control_force.h"

#include "shared/input/joystick.h"

#include "StrExtra.h"
#include "SysParse.h"
#include "FastRandom.h"
#include "SysMalloc.h"

using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

#include "tempstr.h"

#include "../../utils/misc/framerate.h"

/********************************* globals **********************************/

#define PI 3.1415926f

// window/camera stuff
char const AppName[] = "Snow Plow";
int WindowWidth = 1024; //800;
int WindowHeight = 768; // 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 100.0;
Vec2 WorldFromWindowCoords( Vec2 screen );
Vec2 WindowFromWorldCoords( Vec2 world );

GLuint LoadTexture(char *Filename);

float frameTime = 0.0f;
float frameTimeDelta = 1.0f;

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(5.50f, 5.50f);    //a little drag helps with stability
//Gravity_Force gravity(Vec2(0.0f, -1.0f));
Gravity_Force gravity(Vec2(0.0f, 0.0f));
Rigid_Body    worldBody(make_Body_Null(),Vec2(0,0),0,0,0);

char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct GlobalPropertyStruct
{
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;
	
	float GuyMoveMaxVelocity;
	float GuyMoveSpringConstant;

	float GuyJumpScalar;
	float GuyJumpScreenPercent;

	float PowerUpHealth;

	float OutOfWorld;

	int SnowAmount;
	float PlowTurnRate;
	float PlowPushRate;
	float PlowTug;
};

static GlobalPropertyStruct Globals;
static property *globalProperties;
static property_system_editor *GlobalsEditor;

int soundPlay( char *filename, int loop = 0 )
{
	return zsoundPlay( f( "content/sounds/%s.wav", filename ), loop );
}

void soundPlayUnique( int &handle, char *filename, int loop = 0 )
{
	if( handle )
		return;
	handle = zsoundPlay( f( "content/sounds/%s.wav", filename ), loop );
}

void soundStop( int &handle )
{
	if( zsoundIsPlaying( handle ) )
	{
		zsoundStop( handle );
		handle = 0;
	}
}

struct SoundChannel
{
	char name[1024];
	int handle;
	int isLooper;

	SoundChannel( int _loop = 0 )
	{
		name[0] = '\0';
		handle = 0;
		isLooper = _loop;
	}

	void volumeSet( float _volume )
	{
		int loop;
		float volume;
		float pan;
		float freq;
		zsoundGetStats( handle, loop, volume, pan, freq );
		volume = _volume;
		zsoundChange( handle, volume, pan, freq );
	}

	void pitchSet( float _pitch )
	{
		int loop;
		float volume;
		float pan;
		float pitch;
		zsoundGetStats( handle, loop, volume, pan, pitch );
		pitch = _pitch;
		zsoundChange( handle, volume, pan, pitch );
	}

	void play( char *_name )
	{
		if( !stricmp( name, _name ) )
			return;

		strcpy( name, _name );
		soundPlayUnique( handle, name, isLooper );
	}

	void stop()
	{
		soundStop( handle );
		strcpy( name, "" );
	}
};

void KenDrawString( int xPct, int yPct, char *s, int size = 0, int shadow = 0, int center = 0 )
{
	float x = (float)(xPct)*(float)WindowWidth/100.0f;
	float y = (float)(100-yPct)*(float)WindowHeight/100.0f;

	if( center )
		x -= GetTextWidth( size, s ) / 2.0f;

	if( shadow )
	{
		int ofs = size == 0 ? 1 : 3;
		glColor3f( 0.0f, 0.0f, 0.0f );
		DrawString( size, x+ofs, y-ofs, s );
		DrawString( size, x+ofs*2, y-ofs*2, s );
		DrawString( size, x+ofs*-1, y-ofs*-1, s );
	}
	
	glColor3f( 0.3f, 0.30f, 0.9f );
	DrawString( size, x, y, s );
}


class LevelUp;
LevelUp *levelUpMake();
int levelUpProcess( LevelUp *levelUp );
void levelUpKill( LevelUp *levelUp );

class GameFlow
{
public:
	int gameRunning;

	int level;
	int eyeCount;
	int eyeCountClear;
	float timeEyesClear;
	int snowMade;

	int score;
	float bonus;
	int lives;

	int displayState;
	float timeNextDisplay;

	SoundChannel snowMaker;

	LevelUp *levelUp;

	GameFlow() : snowMaker(1)
	{
		gameRunning = 0;
		init();
	}

	void init()
	{
		displayState = 0;
		timeNextDisplay = frameTime + 10.0f;
	
		level = 1;
		score = 0;
		bonus = 0.0f;
		eyeCount = 0;
		eyeCountClear = 0;
		timeEyesClear = 0.0f;
		snowMade = 0;
		lives = 3;
		
		bonusReset();
	}

	void startRunning()
	{
		init();
		gameRunning = 1;
		displayState = 2;
	}

	void stopRunning()
	{
		gameRunning = 0;
		displayState = 3;
		timeNextDisplay = frameTime + 5.0f;
		soundPlay( "truck_breakdown_1" );
	}

	void scoreAdjust( int amount )
	{
		score += amount;
		if( score < 0 )
			score = 0;
	}

	float bonusGetMax()
	{
		return 1000.0f + getMaxEyes() * 100.0f;
	}

	void bonusReset()
	{
		bonus = bonusGetMax();
	}

	void bonusAdjust( float amount )
	{
		bonus += amount;
		if( bonus < 0.0f )
			bonus = 0.0f;
	}

	void levelInc()
	{
		level += 1;
		bonusReset();
	}

	void levelFinish()
	{
		assert( !levelUp );
		levelUp = levelUpMake();
	}
	
	int getSnowCount()
	{
		return Globals.SnowAmount + (int)((Globals.SnowAmount*0.10f)*(float)(level-1));
	}	

	int getMaxEyes()
	{
		return level; // 3+level;
	}

	void startProcessing()
	{
		if( !gameRunning )
		{
			if( frameTime > timeNextDisplay )
			{
				displayState += 1;
				if( displayState > 2 )
					displayState = 0;
				timeNextDisplay = frameTime + 7.0f;
			}
			return;
		}
			
		eyeCountClear = 0;
		if( !levelUp && bonus > 0.0f )
		{
			float mult = 1.0f;
			mult = 1000.0f / 60.0f;	// was 30.0f meaning 30 seconds.
			bonusAdjust( (-frameTimeDelta * mult) );
		}
	}

	void endProcessing();

	void draw()
	{
		if( displayState < 3 )
			KenDrawString( 50, 30, "PITY THE PLOWMAN", 1, 1, 1 );

		if( displayState == 0 )
		{
			KenDrawString( 50, 50, "PRESS SPACE TO START", 1, 1, 1 );
		}
		
		if( displayState ==	1 )
		{
			KenDrawString( 50, 50, "CREDITS", 1, 1, 1 );
			
			int line = 60;
			int x = 30;
			KenDrawString( x, line, "Code & Design - Ken Demarest", 0, 1 );
			line += 4;
			KenDrawString( x, line, "Art - Ryan Ellis", 0, 1 );
			line += 4;
			KenDrawString( x, line, "Art - Daniel Neuburger", 0, 1 );
			line += 4;
			KenDrawString( x, line, "Art - Ryan Junell", 0, 1 );
			line += 4;
			KenDrawString( x, line, "Master Alchemist - Michael Sweet", 0, 1 );
			line += 4;
			KenDrawString( x, line, "Additional Design - Doug Church", 0, 1 );
			line += 4;
			KenDrawString( x, line, "Assistant Flake - Justin Hall", 0, 1 );
			line += 4;
		}

		if( displayState == 2 )
		{
			KenDrawString( 50, 50, "CONTROLS", 1, 1, 1 );

			int line = 60;
			KenDrawString( 20, line, "Use up arrow and down arrow to move", 1, 1 );
			line += 4;
			KenDrawString( 20, line, "Use left and right arrows to turn", 1, 1 );
			line += 4;
			KenDrawString( 20, line, "Use A and D to adjust plow", 1, 1 );
			line += 4;
		}

		if( displayState == 3 )
		{
			KenDrawString( 35, 40, "OUT OF GAS!", 1, 1 );
			KenDrawString( 37, 50, "GAME OVER", 1, 1 );
		}
	}
};


	void GameFlow::endProcessing()
	{
		if( !gameRunning )
			return; 
	
		if( levelUp )
		{
			if( !levelUpProcess( levelUp ) )
			{
				levelUpKill( levelUp );
				levelUp = 0;
			}
		}

		if( eyeCount >= getMaxEyes() && !levelUp )
		{
			if( eyeCountClear >= eyeCount && eyeCount >= 1 )
			{
				timeEyesClear += frameTimeDelta;
			}
			else
			{
				timeEyesClear = 0.0f;
			}

			if( timeEyesClear >= 3.0f && !levelUp )
			{
				levelFinish();
				timeEyesClear = 0.0f;
			}
		}

		if( bonus <= 0.0f && !levelUp )
			stopRunning();
	}

GameFlow *gameFlow = 0;

class LevelUp
{
public:
	int stage;
	float timeNextStage;

	int eyeBonus;
	int gasBonus;
	
	LevelUp()
	{
		stage = 0;
		timeNextStage = frameTime + 2.0f;
		soundPlay( "level_up_1" );
	}

	void stageAdvance()
	{
		stage += 1;
		timeNextStage = frameTime + 1.0f;

		if( stage == 1 )
		{
			eyeBonus = gameFlow->eyeCount * 100;
			gameFlow->scoreAdjust( eyeBonus );
		}

		if( stage == 2 )
		{
			gasBonus = (int)gameFlow->bonus;
			gameFlow->scoreAdjust( (int)gameFlow->bonus );
			gameFlow->bonusReset();
		}

		if( stage <= 3 )
			soundPlay( "score_post_3" );

		if( stage == 5 )
			soundPlay( "newstart" );

	}

	int process()
	{
		if( frameTime >= timeNextStage )
		{
			stageAdvance();
		}
		if( stage >= 7 )
		{
			gameFlow->levelInc();
			return 0;
		}
		return 1;
	}

	void draw()
	{
		int offset = 38;
		int line = 0;
		
		if( stage >= 5 )
		{
			line += 2;
			KenDrawString( 30, offset+(line*6), f( "START LEVEL %d", gameFlow->level+1 ), 1, 1 );
			return;
		}
	
		if( stage >= 0 )
			KenDrawString( 30, offset+(line*6), f( "LEVEL %d COMPLETE!", gameFlow->level ), 1, 1 );
		++line;
		if( stage >= 1 )
			KenDrawString( 30, offset+(line*6), f( "EYE BONUS: +%d", eyeBonus ), 1, 1 );
		++line;
		if( stage >= 2 )
			KenDrawString( 30, offset+(line*6), f( "GAS BONUS: +%d", gasBonus ), 1, 1 );
		++line;
		if( stage >= 3 )
			KenDrawString( 30, offset+(line*6), f( "TOTAL    :  %d", eyeBonus+gasBonus ), 1, 1 );
		++line;
	}
};

LevelUp *levelUpMake()
{
	return new LevelUp;
}

int levelUpProcess( LevelUp *levelUp )
{
	return levelUp->process();
}

void levelUpKill( LevelUp *levelUp )
{
	delete levelUp;
}


float degreesToRadians( float degs )
{
	float rads = PI/180.0f*degs;
	return rads;
}

float radiansToDegrees( float rads )
{
	float degs = 180.0f/PI*rads;
	if( degs < 0.0f )
		degs += 360.0f;
	if( degs > 360.0f )
		degs -= 360.0f;
	return degs;
}

float angleAdd( float angleDegrees, float angleAddDegrees )
{
	float degs = angleDegrees + angleAddDegrees;
	if( degs < 0.0f )
		degs += 360.0f;
	if( degs > 360.0f )
		degs -= 360.0f;
	return degs;
}


float rbGetRadius( Rigid_Body *rb )
{
	Body_Sphere *sphere = dynamic_cast<Body_Sphere*>( rb->m_Pbody.get() );
	if( sphere )
		return sphere->r();

	Body_Box *box = dynamic_cast<Body_Box*>( rb->m_Pbody.get() );
	if( box )
	{
		AABB2 bounds = box->bounds();
		return ( fabs(bounds.x1()-bounds.x0()) + fabs(bounds.y1()-bounds.y0()) ) / 2.0f;
	}

	assert( 0 );
	return 0.0f;
}

//-----------------------------------------------
// Forces
//
typedef Binary_Force Force;
int forceIsName( Force *force, char *nameWild )
{
	return strMatchWild( nameWild, (char*)force->name().c_str() );
}

char *forceGetName( Force *force)
{
	static char temp[1024];
	strcpy( temp, (char*)force->name().c_str() );
	return &(temp[0]);
}

typedef vector<Force*> ForceList;
ForceList forceList;


class Thing;
typedef vector<Thing*> ThingList;

class Hit;
typedef vector<Hit*> HitList;

Thing * thingGetContactAtAngle( Thing *asker, float angleDegrees, float angleToleranceDegrees );
HitList thingGetHitList( Thing *asker );
Thing * thingCreate( char *fileName, Vec2 position, Vec2 velocity );
int thingCountOverlap( Vec2 pos, float radius, char *nameFilter );
ThingList thingGetOverlap( Vec2 pos, float radius, char *nameFilter = 0 );

Force *forceAdd( Thing *thing1, Thing *thing2, float Ks, float Kd, float restLength = 1.0f );
void forceRemove( Force *force );
Force *forceGetByName( char *nameWild );

Thing *rigidBodyToThing( Rigid_Body *rb )
{
	if( !rb )
		return 0;

	return (Thing *)rb->m_pAppData;
}

class Hit
{
public:
	Thing *thing;
	Contact *contact;
	float angleDegrees;

	Hit( Thing *_thing, Contact *_contact )
	{
		thing = _thing;
		contact  = _contact;
	}
};


class Player
{
public:
	float health;
	int backupbeep;
	SoundChannel engine;
	float maxVelocity;
	
	Player() :
		engine(1)
	{
		health = 1.0;
		backupbeep = 0;
		maxVelocity = 1.0f;
	}

	void healthAdjust( float healing )
	{
		health += healing;
		
		if( health < 0.0 )
		{
			health = 0.0f;
		}

		if( health > 1.0  )
		{
			health = 1.0f;
		}
	}

	void harm( float damage )
	{
		healthAdjust( -damage );
	}

	void process( Vec2 position, Vec2 velocity, float pushRate )
	{
		if( !gameFlow->gameRunning )
		{
			engine.stop();
		}
		else
		if( engine.handle == 0 )
		{
			engine.play( "long_2" );
			soundPlay( "truck1" );
		}
	
		maxVelocity = max( velocity.len(), maxVelocity );
		float n = velocity.len() / maxVelocity;
		
//		if( timePushing > 2.0f )
//		{
//			n = 1.0f + ( maxVelocity / velocity.len() );
//		}
		
		engine.pitchSet( 1.0f + n*10.0f );
		engine.volumeSet( 0.75f + (n*0.25f) );
	}
	
};

void creaturePick( char *name )
{
	switch( fastRandomInt(4) )
	{
		case 0:
			strcpy( name, "beetle" );
			break;
		case 1:
			strcpy( name, "spider" );
			break;
		case 2:
			strcpy( name, "fly" );
			break;
		case 3:
			strcpy( name, "ladybug" );
			break;
	}
}

class Generator
{
public:
	Thing *parent;
	double nextGeneration;
	double period;
	int countMade;
	int countToMake;

	Generator( Thing *_parent, float _period, int _count )
	{
		assert( _parent );
		parent = _parent;
		period = _period;
		nextGeneration = frameTime + 0.0f; //period;
		countToMake = _count;
		countMade = 0;
	}

	~Generator()
	{
		gameFlow->snowMaker.stop();
	}
	
	int readyToGenerate()
	{
		return frameTime > nextGeneration && countMade < countToMake;
	}
	
	int doneMaking()
	{
		return countMade >= countToMake;
	}
	
	void generate( char *s, Vec2 position, Vec2 velocity )
	{
		char creatureName[1024];
		strcpy( creatureName, s );
		if( !stricmp( creatureName, "any" ) )
		{
			creaturePick( creatureName );	
		}

		if( !stricmp( s, "snow" ) )
		{
			gameFlow->snowMaker.play( "snow_maker_loop_1" );
			gameFlow->snowMaker.volumeSet( 0.85f );
			gameFlow->snowMade += 1;
		}
	
		float rad = fastRandomFloat(PI*2.0f);
		position.m_v[0] += cos(rad) * 0.5f;
		position.m_v[1] += sin(rad) * 0.5f;

		thingCreate( creatureName, position, velocity );

		countMade += 1;
	}

	void testGeneration( char *myName, Vec2 position, Vec2 velocity )
	{
		if( readyToGenerate() )
		{
			int reps = 1;
			while( reps > 0 && countMade < countToMake )
			{
				char temp[1024];
				char *s = sysParse( 1024, temp, myName, "_" );
				generate( s, position, velocity );
				nextGeneration = frameTime + period;
				--reps;
			}
		}
	}
};


class Tongue
{
public:
	Thing *parent;
	Thing *thingAttached;
	Force *forceAttaching;
	double timeLastUnstuck;
	float suctionStrength;
	
	Tongue( Thing *_parent )
	{
		assert( _parent );
		parent = _parent;
		thingAttached = 0;
		forceAttaching = 0;
		timeLastUnstuck = 0.0f;
		suctionStrength = 0.0f;
	}
	int isStuck()
	{
		return thingAttached != 0;
	}
	int isSticky()
	{
		if( frameTime < timeLastUnstuck + 1.0 )
			return 0;
		return 1;
	}
	void stickTo( Thing *_thing, Hit *_hit, float _strength )
	{
		assert( thingAttached == 0 );
		assert( forceAttaching == 0 );
		assert( frameTime >= timeLastUnstuck + 1.0f );
		
		suctionStrength = _strength;
		
		thingAttached = _thing;
		
		forceAttaching = forceAdd( parent, thingAttached, 100.0f, 20.0f, 0.0f );
		assert( forceAttaching );
	}

	void unStick()
	{
		assert( forceAttaching );
		forceRemove( forceAttaching );
		delete forceAttaching;
		forceAttaching = 0;
		thingAttached = 0;
		suctionStrength = 0.0f;
		
		timeLastUnstuck = frameTime;
	}

	int shouldLoosen( float externalForces )
	{
		if( externalForces > suctionStrength )
		{
			return 1;
		}
		return 0;
	}
};

/*
Creature types
	Fly - flies quickly, eratically, keeps flying if you don't hit it solidly
		  enough with your tongue to knock it out.  Gets away in short time unless
		  you've stunned it.
	Beetle - slow, heavy, never gets away, but its weight can cause your suction
		  to not be strong enough.
	Spider - medium sized, medium mover, if it touches your head then it stings
		  you and harms you. When it stings you let go.
	Ladybug - benign, medium speed eratic mover, but very small.


*/
PROP_STRUCT struct CreaturePropertyStruct
{
	float points;
	float speed;
	float leanSwitch;
	float turnRateMax;
};

PROP_STRUCT struct TexturePropertyStruct
{
	float scale;
	float z;
};

float roll( float occurEveryXSeconds )
{
	return( fastRandomFloat( occurEveryXSeconds ) < frameTimeDelta );
}


class Creature
{
public:
	Thing *parent;
	CreaturePropertyStruct props;

	float lean;

	Creature( Thing *_parent, char *creatureName )
	{
		assert( _parent );
		parent = _parent;
		
		property *templateGroup = FindSubPropertyByName( LocalTemplates, "CreaturePropertyStruct" );
		property *masterGroup   = FindSubPropertyByName( globalProperties, "CreatureData" );
		property *dataGroup     = FindSubPropertyByName( masterGroup, creatureName );
		ConvertPropertyToStruct( templateGroup, dataGroup, &props );

		lean = 1.0f;
	}

	float getSpeed()
	{
		return props.speed;
	}
	
	float getPoints()
	{
		return props.points;
	}

	void move( Vec2 &velocity, float &heading )
	{
		if( props.speed > 0.0f )
		{
			if( roll(props.leanSwitch) )
				lean = -lean;
				
			if( velocity.len() < props.speed )
			{
				velocity.m_v[0] += cos( heading );
				velocity.m_v[1] += sin( heading );
			}
			float turnAmount = PI/2.0f / (props.turnRateMax / frameTimeDelta);
			heading += turnAmount * lean;
		}
	}
};

class Score
{
public:
	float score;
	Score()
	{
		score = 0.0f;
	}
	void scoreAdjust( float points )
	{
		score += points;
	}
};


class Eye
{
public:
	int everMadeNoise;
	int wasClear;
	int clear;
	
	Eye()
	{
		clear = 0;
		wasClear = 0;
	}

	void process( Rigid_Body *rb )
	{
		int snowCover = thingCountOverlap( rb->x, rbGetRadius(rb), "snow" );
		
		if( snowCover == 0 )
		{
			clear = 1;		
			gameFlow->eyeCountClear += 1;
		}
		else
		{
			clear = 0;
		}
		if( clear != wasClear )
		{
			if( clear )
				soundPlay( "eye_open" );
			else
				soundPlay( "eye_close" );
		}
		wasClear = clear;	

		if( !everMadeNoise )
		{
			soundPlay( "eye_appear" );
			everMadeNoise = 1;
		}
	}
};

class TextureAnim
{
public:
	GLuint texture;
	int isVisible;
	float lastChange;
	int frame;
	float scale;
	float scaleMag;

	TextureAnim()
	{
		isVisible = 1;
		texture = 0;
		frame = 0;
		lastChange = frameTime;
		scale = 1.0f;
		scaleMag = 1.0f;
	}

	int isEmpty()
	{
		return texture == 0;
	}

	int get()
	{
		return texture;
	}

	void load( char *fileName )
	{
		texture  = LoadTexture( fileName );
	}

	void change( int dir, float speed )
	{
		if( frameTime > lastChange + speed )
		{
			frame += dir;
			lastChange = frameTime;
		}
	}

};

TextureAnim backgroundTexture;

//-----------------------------------------------
// Thing
//
class Thing
{
public:
    Rigid_Body *rb;
	Player *    player;
	Generator * generator;
	Tongue *    tongue;
	Creature *  creature;
	Score *     score;
	Eye *		eye;

	TexturePropertyStruct textureProps;
	TextureAnim textureAnim;

	float jumpAngle;
	float jumpMagnitude;

	float turnRate;
	float pushRate;

	float health;

	float timeCreated;
	int alive;
    
	Thing( Rigid_Body *_rb )
	{
		alive = 1;
		player = 0;
		generator = 0;
		tongue = 0;
		creature = 0;
		score = 0;
		eye = 0;

		jumpAngle = 0.0f;
		jumpMagnitude = 0.0f;
		turnRate = 0.0f;
		pushRate = 0.0f;

		timeCreated	= frameTime;
		
		rb = _rb;
		assert(rb);
		rb->m_pAppData = this;

		property *templateGroup = FindSubPropertyByName( LocalTemplates, "TexturePropertyStruct" );
		property *masterGroup   = FindSubPropertyByName( globalProperties, "TextureData" );
		property *dataGroup     = FindSubPropertyByName( masterGroup, getName() );
		ConvertPropertyToStruct( templateGroup, dataGroup, &textureProps );
		if( textureProps.scale == 0.0f )
			textureProps.scale = 1.0f;

		textureAnim.load( getName() );
		textureAnim.scale = textureProps.scale;

		if( isName( "player*" ) )
			player = new Player;

		if( isName( "gen_*" ) )
		{
			generator = new Generator( this, 0.01f, Globals.SnowAmount );
			textureAnim.isVisible = 0;
		}

		if( isName( "tongue*" ) )
			tongue = new Tongue( this );

		if( isName( "score*" ) )
			score = new Score();

		if( isName( "eye*" ) )
		{
			eye = new Eye();
			textureAnim.scaleMag = 10.0f;
			soundPlay( "win_level" );
		}

		if( isCreature() )
			creature = new Creature( this, getName() );
		
		simulator.add(*rb);
	}

    ~Thing()
    {
		simulator.remove(*rb);
    
		if( player ) delete player;
		if( generator ) delete generator;
		if( tongue ) delete tongue;
		if( creature ) delete creature;
		if( score ) delete score;
		if( eye ) delete eye;
		delete rb;
		rb = 0;
    }

	float getRadius()
	{
		return rbGetRadius( rb );
	}

	int isDead()
	{
		return !alive;
	}

	int isName( char *nameWild )
	{
		assert( rb );
		return strMatchWild( nameWild, (char*)rb->name().c_str() );
	}

	char *getName()
	{
		static char temp[1024];
		assert( rb );
		strcpy( temp, (char*)rb->name().c_str() );
		return &(temp[0]);
	}

	// The rate is how many seconds it would take you to turn 360 degrees.
	void turn( float rate )
	{
		if( rate == 0.0f )
			return;
		float &heading = rb->O.m_z;
		float f = frameTimeDelta;
		if( f > 1.0f / 30.0f )
			f = 1.0f / 30.0f;
		heading += degreesToRadians( ( 360.0f / rate ) * f );
		rb->update_derived_state();
	}

	// The rate is how many seconds it would take you to turn 360 degrees.
	void setTurn( float rate );

	// The rate is how many 
	void push( float rate )
	{
		float &heading = rb->O.m_z;
		Vec2 &velocity = rb->v;
		Vec2 strength( cos( heading )*rate, sin( heading )*rate );
		float f = frameTimeDelta;
		if( f > 1.0f / 30.0f )
			f = 1.0f / 30.0f;
		velocity += strength * f;
		rb->update_derived_state();
	}

	void setPush( float rate )
	{
		pushRate = rate;
	}

	void jumpApply( float angleDegrees, float magnitude )
	{
		float angleOppositeDegrees = angleAdd( angleDegrees, 180.0f );
		Thing *floor = thingGetContactAtAngle( this, angleOppositeDegrees, 180.0f );
		if( floor )
		{
			float angleRadians = degreesToRadians( angleDegrees );
			Vec2 jumpVector( cos( angleRadians ), sin( angleRadians ) );
			jumpVector *= magnitude;
			rb->v = jumpVector;
		}
	}

	void analyzeHits()
	{
		HitList hitList = thingGetHitList( this );
		for( int i=0 ; i<hitList.size() ; ++i )
		{
			Hit *hit = hitList[i];
			
			if( hit->thing->isDead() )
				continue;
				
			if( hit->thing->isName( "gift*" ) )
			{
				gameFlow->bonusAdjust( 500 );
				ThingList meltList = thingGetOverlap( hit->thing->rb->x, hit->thing->getRadius() * 4, "snow*" );
				for( int i=0 ; i<meltList.size() ; ++i )
					meltList[i]->die();

				soundPlay( "bonus_2" );
				hit->thing->die();
			}
			if( hit->thing->isName( "mine*" ) )
			{
				hit->thing->die();
				gameFlow->scoreAdjust( -1000 );
				Thing *boom = thingCreate( "boom", hit->thing->rb->x, Vec2(0.0f,0.0f) );
				soundPlay( "explosion_2" );
				extern void mineMake();
				mineMake();
			}
		}
	}

	float setPlow( char *name, float base, float tug );



	void tryUnStick()
	{
		if( tongue && tongue->isStuck() )
		{
			assert( tongue );
			tongue->unStick();
			assert( tongue->thingAttached == 0 );
		}
	}

	int isCreature()
	{
		return isName( "gem*" ) || isName( "spider*" ) || isName( "beetle*" ) || isName( "fly*" ) || isName( "ladybug*" );
	}

	void trySticking()
	{
		assert( tongue );
		if( tongue->isStuck() || !tongue->isSticky() )
			return;
			
		HitList hitList = thingGetHitList( this );
		for( int i=0 ; i<hitList.size() ; ++i )
		{
			Hit *hit = hitList[i];
				
			if( hit->thing->isCreature() )
			{
				Vec2 vecSum = rb->x + hit->thing->rb->x;
			
				tongue->stickTo( hit->thing, hit, vecSum.len() );
				return;
			}
		}
	}

	void scorePoints()
	{
		assert( score );
		
		HitList hitList = thingGetHitList( this );
		for( int i=0 ; i<hitList.size() ; ++i )
		{
			Hit *hit = hitList[i];
			if( hit->thing->isCreature() && !hit->thing->isDead() )
			{
				score->scoreAdjust( hit->thing->creature->getPoints() );
				hit->thing->die();
				
			}
		}			
			
	}

	void die()
	{
		alive = 0;
	}

	void checkOutOfWorld()
	{
		Vec2 dist = rb->x;
		if( dist.len() > Globals.OutOfWorld )
			die();
	}
	
	void process()
	{
		if( isDead() )
			return;
			
		if( jumpMagnitude > 0.0f )
		{
			jumpApply( jumpAngle, jumpMagnitude );
			jumpMagnitude = 0.0f;
		}

		if( isName( "plow*" ) )
		{
			analyzeHits();
		}

		if( isName( "mine*" ) )
		{
			int n = (int)( (frameTime - timeCreated)/2.0f );
			textureAnim.frame = n % 2;
			if( textureAnim.frame == 0 )
				textureAnim.load( "mine0" );
			else
				textureAnim.load( "mine1" );
		}


		if( isName( "fuel_needle" ) )
		{
			float howFull = gameFlow->bonus / gameFlow->bonusGetMax();
			if( howFull > 1.0f )
				howFull = 1.0f;
				
			float degreeRange = 80.0f;
			float degreeStart = 360.0f + (degreeRange/2.0f);
			float rads = degreesToRadians( degreeStart - (degreeRange*howFull) );
			float &heading = rb->O.m_z;
			
			heading = rads;
		}

		if( player )
		{
			analyzeHits();
			if( !gameFlow->gameRunning )
			{
				static float timeNextAct = 0.0f;
				if( frameTime > timeNextAct )
				{
					turnRate = 0.0f;
					pushRate = 0.0f;
					int n = fastRandomInt( 100 );
					if( n < 10 )
						turnRate = Globals.PlowTurnRate * 0.25f + fastRandomFloat( Globals.PlowTurnRate * 0.75f );
					else
					if( n < 20 )
						turnRate = - ( Globals.PlowTurnRate * 0.25f + fastRandomFloat( Globals.PlowTurnRate * 0.75f ) );
					else
					if( n < 90 )
						pushRate = Globals.PlowPushRate;
					else
						pushRate = -Globals.PlowPushRate;
					timeNextAct = frameTime + 0.2f + fastRandomFloat( 2.0f );
				}
			}
			
			turn( turnRate );
			push( pushRate );

			player->process( rb->x, rb->v, pushRate );
		}
		
		if( generator )
		{
			generator->testGeneration( getName(), rb->x, rb->v );
			if( generator->doneMaking() )
				die();
		}
		
		if( tongue )
		{
			trySticking();

			if( tongue->isStuck() )
			{
				Vec2 vecSum = tongue->thingAttached->rb->x;
				if( tongue->shouldLoosen( vecSum.len() ) )
					tryUnStick();
			}
		}

		if( creature )
		{
			creature->move( rb->v, rb->O.m_z );
		}

		if( score )
			scorePoints();

		if( eye )
		{
			if( textureAnim.scaleMag > 1.0f )
			{
				textureAnim.scaleMag -= frameTimeDelta * 4.0f;
				if( textureAnim.scaleMag <= 1.0f )
					textureAnim.scaleMag = 1.0f;
			}
				
			eye->process( rb );
			if( eye->clear && textureAnim.frame > 0 )
				textureAnim.change( -1, 0.2f );
			if( !eye->clear && textureAnim.frame < 5 )
				textureAnim.change( 1, 0.2f );
			textureAnim.load( f( "eye%d", textureAnim.frame ) );
		}

		if( isName( "boom" ) )
		{
			if( textureAnim.frame < 60 )
			{
				textureAnim.frame += 1;
				textureAnim.load( f( "explode%04d", textureAnim.frame ) );
			}
			else
			{
				if( frameTime - timeCreated > 0.2f )
					die();
			}
		}

		checkOutOfWorld();
	}

};


	void Thing::setTurn( float rate )
	{
		turnRate = rate;
		assert( rate <= 5.0f );
		static float turnRateMax;
		static float turnRateMin;
		if( rate > turnRateMax )
			turnRateMax = rate;
		if( rate < turnRateMin )
			turnRateMin = rate;
	}



	float Thing::setPlow( char *name, float base, float tug )
	{
		Force *force = forceGetByName( name );
		assert( force );
		if( force )
		{
			Spring_Force *sf = (Spring_Force *)force;
			if( base == 0.0f )
				base = sf->m_rest_len;
			sf->m_rest_len = base * tug;
		}
		return base;
	}


ThingList thingList;

ThingList thingGetOverlap( Vec2 pos, float radius, char *nameFilter )
{
	ThingList overlapList;
	for( int i=0 ; i<thingList.size() ; ++i )
	{
		Thing *thing = thingList[i];
		if( !nameFilter || nameFilter[0] == '\0' || thing->isName( nameFilter ) )
		{
			float myRadius = rbGetRadius( thing->rb );
			Vec2 myPos = thing->rb->x;
			
			Vec2 delta = pos - myPos;
			if( delta.len() <= radius+myRadius )
			{
				overlapList.push_back( thing );
			}
		}
	}

	return overlapList;
}

int thingCountOverlap( const Vec2 pos, float radius, char *nameFilter = 0 )
{
	ThingList overlapList = thingGetOverlap( pos, radius, nameFilter );
	return (int)overlapList.size();
}


ThingList thingGetByNameWild( char *nameWild )
{
	ThingList list;
	for( int i=0 ; i<thingList.size() ; ++i )
	{
		Thing *thing = thingList[i];
		assert( thing );
		if( thing->isName( nameWild ) )
		{
			list.push_back( thing );
		}
	}
	return list;
}

Thing *thingGetByName( char *nameWild )
{
	for( int i=0 ; i<thingList.size() ; ++i )
	{
		Thing *thing = thingList[i];
		assert( thing );
		if( thing->isName( nameWild ) )
		{
			return thing;
		}
	}
	return 0;
}


Force *forceGetByName( char *nameWild )
{
	for( int i=0 ; i<forceList.size() ; ++i )
	{
		Force *force = forceList[i];
		assert( force );
		if( forceIsName( force, nameWild ) )
		{
			return force;
		}
	}
	return 0;
}


Thing *thingGetContactAtAngle( Thing *asker, float angleDegrees, float angleToleranceDegrees )
{
	for( Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); ++cit )
	{
        Rigid_Body *r1 = (Rigid_Body *)(cit->m_Prb_a);
        Rigid_Body *r2 = (Rigid_Body *)(cit->m_Prb_b);
        if( ( r1 == asker->rb || r2 == asker->rb ) && r1 && r2 )
        {
			Vec2 contactPointInBodySpace = asker->rb->world_space_to_body_space( cit->m_x );
			float contactAngleRadians = atan2( contactPointInBodySpace.m_v[1], contactPointInBodySpace.m_v[0] );
			float contactAngleDegrees = radiansToDegrees( contactAngleRadians );
			float delta = fabs( angleDegrees - contactAngleDegrees );
			if( delta < angleToleranceDegrees || delta > 360 - angleToleranceDegrees )
			{
				if( asker->rb == r1 )
					return (Thing *)r2->m_pAppData;
				else
					return (Thing *)r1->m_pAppData;
			}
		}
	}
	return 0;
}

//
// Warning: does not return 'dead' things.
//
HitList thingGetHitList( Thing *asker )
{
	HitList hitList;
	
	for( Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); ++cit )
	{
        Rigid_Body *r1 = (Rigid_Body *)(cit->m_Prb_a);
        Rigid_Body *r2 = (Rigid_Body *)(cit->m_Prb_b);
        if( ( r1 == asker->rb || r2 == asker->rb ) && r1 && r2 )
        {
			Thing *thingHit = ( asker->rb == r1 ) ? rigidBodyToThing(r2) : rigidBodyToThing(r1);
			if( !thingHit->isDead() )
			{
				hitList.push_back( new Hit( thingHit, cit._Myptr ) ); //cit ) );
			}
		}
	}
	return hitList;
}

class Board
{
public:
	int initialized;	
	Vec2 topLeft;
	Vec2 bottomRight;

	double timeNextGenerate;
	
	Board()
	{
		initialized	= 0;
	}

	void init()
	{
		Thing *thing = thingGetByName( "corner_tl" );	
		assert( thing );
		topLeft = thing->rb->x;
		
		thing = thingGetByName( "corner_br" );	
		assert( thing );
		bottomRight = thing->rb->x;
	}

	float getWidth()
	{
		return bottomRight.m_v[0] - topLeft.m_v[0];
	}

	float getHeight()
	{
		return topLeft.m_v[1] - bottomRight.m_v[1];
	}

	Vec2 posGetRandom( float margin )
	{
		Vec2 pos;
		float xMargin = getWidth() * margin;
		float yMargin = getHeight() * margin;
		
		pos.m_v[0] = topLeft.m_v[0] + xMargin + fastRandomFloat( getWidth() - (xMargin*2.0f) );
		pos.m_v[1] = bottomRight.m_v[1] + yMargin + fastRandomFloat( getHeight() - (yMargin*2.0f) );

		return pos;
	}
	
	void eyeMake()
	{
		Thing *eye = thingCreate( "eye", Vec2(0.0f,0.0f), Vec2(0.0f,0.0f) );
		assert( eye );

		int eyeOverlap = 0;
		int mineOverlap = 0;
		int snowOverlap = 0;
		do {
			eye->rb->x = posGetRandom( 0.03f );
			eye->rb->update_derived_state();
			eyeOverlap =  thingCountOverlap( eye->rb->x, eye->getRadius(), "eye" );
			mineOverlap =  thingCountOverlap( eye->rb->x, eye->getRadius(), "mine" );
			snowOverlap = thingCountOverlap( eye->rb->x, eye->getRadius(), "snow" );
		} while( eyeOverlap > 1 || snowOverlap <= 1 || mineOverlap > 0 );

		gameFlow->eyeCount += 1;
	}

	void mineMake()
	{
		Thing *mine = thingCreate( "mine", Vec2(0.0f,0.0f), Vec2(0.0f,0.0f) );
		assert( mine );

		int eyeOverlap = 0;
		int mineOverlap = 0;
		int playerOverlap = 0;
		do {
			mine->rb->x = posGetRandom( 0.10f );
			mine->rb->update_derived_state();
			eyeOverlap =  thingCountOverlap( mine->rb->x, mine->getRadius(), "eye" );
			mineOverlap =  thingCountOverlap( mine->rb->x, mine->getRadius(), "mine" );
			playerOverlap =  thingCountOverlap( mine->rb->x, mine->getRadius()*4.0f, "player" );
		} while( eyeOverlap > 0 || mineOverlap > 1 || playerOverlap > 0 );
	}

	void giftMake()
	{
		Thing *gift = thingCreate( "gift", Vec2(0.0f,0.0f), Vec2(0.0f,0.0f) );
		assert( gift );

		int eyeOverlap = 0;
		do {
			gift->rb->x = posGetRandom( 0.10f );
			gift->rb->update_derived_state();
			eyeOverlap =  thingCountOverlap( gift->rb->x, gift->getRadius(), "eye" );
		} while( eyeOverlap > 0 );
	}


	void process()
	{
		if( !initialized )
			init();

		
		if( gameFlow->snowMade > 100 && !gameFlow->levelUp && gameFlow->eyeCount < gameFlow->getMaxEyes() )
		{
			eyeMake();
			if( gameFlow->eyeCount > 1 )
			{
				mineMake();
				giftMake();
			}
		}
	}
};


static Vec2 mouseDown,mouseScreen,mouseWorld;
static Joystick PS2Controllers;
static Thing *mouseThing = 0;
static Force *mouseSpring = 0;
static Board *board = 0;

void mineMake()
{
	if( board )
		board->mineMake();
}

struct TextureCache
{
	char name[80];
	int  texture;
};	

typedef vector<TextureCache*> TextureCacheList;
TextureCacheList textureCacheList;

int textureCacheGetByName( char *name )
{
	for( int i=0 ; i<textureCacheList.size() ; ++i )
	{
		if( !stricmp( textureCacheList[i]->name, name ) )
			return textureCacheList[i]->texture;
	}
	return 0;
}

GLuint LoadTexture(char *fileName)
{
    GLuint texture = textureCacheGetByName( fileName );
    if( texture )
		return texture;
    
	f path( "content\\bitmaps\\%s.psd", fileName );
    
    Bitmap_Load_Result result;
    int ok = LoadBitmapByExtension( path.s, &result );
    
    if( !ok )
		return 0;

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

	TextureCache *textureCache = new TextureCache;
	textureCache->texture = texture;
	strcpy( textureCache->name, fileName );
	textureCacheList.push_back( textureCache );

    return textureCache->texture;
}

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Force *force)
{
    if(force)
    {
        glBegin(GL_LINES);
        glVertex2f(force->m_a.x_worldS().x(), force->m_a.x_worldS().y());
        glVertex2f(force->m_b.x_worldS().x(), force->m_b.x_worldS().y());
        glEnd();
    }
}

Force *forceAdd( Thing *thing1, Thing *thing2, float Ks, float Kd, float restLength )
{

    Force *f = new Spring_Force(
        Handle( *thing1->rb, Vec2(0.0f,0.0f) ),
        Handle( *thing2->rb, Vec2(0.0f,0.0f) ), 
        Ks,
        Kd,
        restLength
        );
    forceList.push_back( f );
    simulator.add( *f );
	return f;
}

void forceRemove( Force *force )
{
	ForceList::iterator bfi = std::find( forceList.begin(), forceList.end(), force );
	forceList.erase( bfi );
	simulator.remove( *force );
}


/*
void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}
*/

void ClearAll()
{
//    ClearMouseSpring();
    
    for(int i = 0;i < thingList.size();++i)
    {
        assert(thingList[i]);
        simulator.remove(*thingList[i]->rb);
    }

    for(int i = 0;i < forceList.size();++i)
    {
        assert(forceList[i]);
        simulator.remove(*forceList[i]);
    }

    thingList.clear();
    forceList.clear();
}

Thing *thingCreate( char *fileName, Vec2 position, Vec2 velocity )
{
	Thing *thingCreated = 0;

    int num_rb=0, num_force=0;
    Rigid_Body **rb_list = 0;
    Force **force_list = 0;

	f path( "content\\critters\\%s.physdb", fileName );

    loadPhysicsFile(path.s, 0, 0, 1.0f/10.0f, 1.0f/16.0f,
                    &rb_list, &num_rb, &force_list, &num_force, &worldBody);

    for (int i=0; i < num_rb; ++i)
    {
		rb_list[i]->x = position;
		rb_list[i]->v = velocity;
		
        Thing *addMe = new Thing( rb_list[i] );
		addMe->rb->x = position;
		addMe->rb->v = velocity;
        if( i == 0 )
        {
			thingCreated = addMe;
		}
        thingList.push_back( addMe );
    }

    for (int i=0; i < num_force; ++i)
    {
        forceList.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }
    
	return thingCreated;
}

void
LoadTestFile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Force **force_list;

    loadPhysicsFile(TestLevelFileName, 0, 0, 1.0f/10.0f, 1.0f/16.0f,
                    &rb_list, &num_rb, &force_list, &num_force, &worldBody);

    for (int i=0; i < num_rb; ++i)
    {
		Thing *made = new Thing( rb_list[i] );
		thingList.push_back( made );
    }

    for (int i=0; i < num_force; ++i)
    {
        forceList.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }
}


/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
	sysMallocInit( 0, 0, 0, 0, 0, 0 );


	fastRandomize();
    InitializeFrameRate(30.0f);
	PS2Controllers.init();

/*
    Globals.Gravity = 0.0f; //-40.0f;

    Globals.MouseSpringKs = 100.0f;
    Globals.MouseSpringKd = 20.0f;
    Globals.MouseSpringRestLength = 0.0f;

    Globals.GuyMoveMaxVelocity    = 40.0f;
    Globals.GuyMoveSpringConstant = 30.0f;

	Globals.GuyJumpScreenPercent  = 0.50f;	// Mouse over 20% of the screen for a 100% jump.
	Globals.GuyJumpScalar         = fabs(Globals.Gravity) * 4.0f;	// Your boost at a 100% jump.

	Globals.PowerUpHealth         = 0.2f;
	Globals.OutOfWorld			  = 200.0f;

	Globals.SnowAmount			  = 300;
	Globals.PlowTurnRate		= 5.0f;
	Globals.PlowPushRate		= 10.0f;
*/
	
    globalProperties = ReadPropertyFromFile( "default.igj2props" );

    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, globalProperties);

    ConvertPropertyToStruct(
		// This is the struct in this file to find, by struct type.
		FindSubPropertyByName(LocalTemplates, "GlobalPropertyStruct"),
		// This is the actual properties inside the master list
        FindSubPropertyByName(globalProperties,"GlobalPropertyData"),
        &Globals);
	
    
    frameTime = (float)GetTime(); // throw away initial 0.0 so we're seconds-from-startup
	gameFlow = new GameFlow;
	gameFlow->init();

    // gl stuff
    glClearColor(0.0f,0.0f,0.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
//    music.loadseg("background", "Segment1.sgp", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);


    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(worldBody);

    LoadTestFile();

	board = new Board;
}

void RestartGame()
{
	ClearAll();
	delete board;
	board = new Board;
	LoadTestFile();
	gameFlow->startRunning();
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{

	assert( gameFlow );
	assert( board );
	
    float newFrameTime = (float)GetTime(); // throw away initial 0.0 so we're seconds-from-startup
    frameTimeDelta = newFrameTime - frameTime;
    frameTime = newFrameTime;

	gameFlow->startProcessing();

    ConvertPropertyToStruct(
		// This is the struct in this file to find, by struct type.
		FindSubPropertyByName(LocalTemplates, "GlobalPropertyStruct"),
		// This is the actual properties inside the master list
        FindSubPropertyByName(globalProperties,"GlobalPropertyData"),
        &Globals);


	ThingList killList;
	for( int i=0 ; i<thingList.size() ; i++ )
	{
		thingList[i]->process();
			
		if( thingList[i]->isDead() )
		{
			Thing *killMe = thingList[i];
			killList.push_back( killMe );
		}

		Thing *thingSource = thingList[i];
		if( thingSource->isName( "make_*" ) )
		{
			char name[1024];
			strcpy( name, thingSource->getName() );
			char dummy[1024];
			char *nameToMake = sysParse( 1024, dummy, name, "_" );
			thingCreate( nameToMake, thingSource->rb->x, thingSource->rb->v );
			if( !thingSource->isDead() )
			{
				killList.push_back( thingSource );
			}
		}
	}

	for( int i=0 ; i<killList.size() ; i++ )
	{
		Thing *thingKill = killList[i];
		delete thingKill;
		thingList.erase( std::find( thingList.begin(), thingList.end(), thingKill ) );
	}

	board->process();
	
    gravity.m_g.set_y(Globals.Gravity);
    
    /**************************** step the simulator ****************************/   

	PS2Controllers.sample();

    if( mouseSpring )                             //update
        mouseSpring->m_a = Handle(worldBody, mouseWorld);

    int const NumberOfSimSteps = 1;
    for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
        // step as many times as you like
        simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
        // do any game stuff you want here (like per-step collision response, etc.)
    }
    // now wait for the rest of the time so we're consistent on fast machines
    WaitFrameRate(GetFrameRate());

	if( gameFlow->gameRunning )
	{
		for (int i=0; i<PS2Controllers.num_joysticks(); i++)
		{
			Thing *player = thingGetByName( "player*" );
			if( player )
			{
				float pushRate = PS2Controllers.right_stick(i).y() * Globals.PlowPushRate;
				player->setPush( pushRate );

				float raw = PS2Controllers.right_stick(i).x();
/*
				if( raw > 1.0f )
					raw = 1.0f;
				if( raw < -1.0f )
					raw = -1.0f;
				float margin = 0.25f;
				if( raw > margin )
					raw -= margin;
				if( raw < -margin )
					raw += margin;
				raw = raw * (1.0f/(1.0f-margin));
*/
				if( raw >= -0.5f && raw <= 0.5f )
					raw = 0.0f;
				float turnRate = -raw*Globals.PlowTurnRate;
				assert( turnRate <= Globals.PlowPushRate );
				assert( turnRate >= -Globals.PlowPushRate );
				player->setTurn( turnRate );
			}
		}
	}
	else
	{
		for (int i=0; i<4; i++)
		{
			if (PS2Controllers.button_went_down(Joystick::BUTTON_RIGHT_U, i))
			{
				//ResetGame();
				break;
			}
		}
	}


    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glColor3f(0,0,0);
	glDepthMask( GL_FALSE );

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
/*
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();
*/

    // draw mouse
/*
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();
*/

//    glForce1v(mouseSpring); 

    vector<Vec2> textureCoords;
	textureCoords.push_back(Vec2( 1, 0));
	textureCoords.push_back(Vec2( 0, 0));
    textureCoords.push_back(Vec2( 0, 1));
	textureCoords.push_back(Vec2( 1, 1));

    vector<Vec2> bodyCoords;
	bodyCoords.push_back(Vec2( 0.5, 0.5));
    bodyCoords.push_back(Vec2(-0.5, 0.5));
	bodyCoords.push_back(Vec2(-0.5,-0.5));
	bodyCoords.push_back(Vec2( 0.5,-0.5));
	
	// iterate all the bodies, rendering them

	int done = 0;
	float drawDepth    = 0.0f;
	float alreadyDrawn = -100.0f;
	while( !done )
	{
		drawDepth = 100.0f;
	    for( int i = 0;i < thingList.size(); ++i )
	    {
	        Thing *thing = thingList[i];
			if( thing->textureProps.z < drawDepth && thing->textureProps.z > alreadyDrawn )
			{
				drawDepth = thing->textureProps.z;
			}
	    }
	    if( drawDepth == 100.0f )
			break;
	    
		for(int i = 0;i < thingList.size();++i)
		{
			assert(thingList[i]);
			Thing &thing = *thingList[i];

			if( thing.textureProps.z == drawDepth && thing.textureAnim.isEmpty() && thing.textureAnim.isVisible )
			{
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				glColor3f(0,0,1);
				//glBody1v(thing.rb->m_Pbody.get(), 1);
			}
			
			if( thing.textureProps.z == drawDepth && !thing.textureAnim.isEmpty() && thing.textureAnim.isVisible )
			{
				float scale = 1.0f;
				scale = thing.textureAnim.scale * thing.textureAnim.scaleMag;
			
				glEnable(GL_TEXTURE_2D);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				{
					glBindTexture( GL_TEXTURE_2D, thing.textureAnim.texture );
					glPushMatrix();
			        
					float z = thing.textureProps.z;
					
					glTranslatef( thing.rb->x.m_v[0], thing.rb->x.m_v[1], z );
					glRotatef( radiansToDegrees( thing.rb->O.m_z ), 0.0f, 0.0f, 1.0f );
					glColor4f(1,1,1,1.0f);

					glBegin(GL_POLYGON);
			      
					for( int j = 0; j < textureCoords.size();++j )
					{
						// get out our body-space rendering vertex
						Vec2 tc = textureCoords[j];
						Vec2 bc = bodyCoords[j];
						// transform it by the body's pos/rot
						Vec2 p = bc * scale; //+ thing.rb->R * bc;

						glTexCoord2f( tc.x(), tc.y() );
						glVertex2f( p.x(), p.y() );
					}
					glEnd();
					glPopMatrix();
				}
				glDisable(GL_TEXTURE_2D);
				glDisable(GL_BLEND);

				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

				// iterate all the forces, rendering them
				//glColor3f(1,0,0);
				//glBody1v(thing.rb->m_Pbody.get(), 1);
			}
		}
		alreadyDrawn = drawDepth;
	}
    
    glDisable(GL_BLEND);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor3f(1,0,0);
    for(int i = 0;i < forceList.size();++i)
    {
        assert(forceList[i]);
        Force *force = forceList[i];

        glForce1v(force);
    }

	glColor3f(0,0,1);
    glDisable(GL_BLEND);

/*
    glColor3f(0, 0, 0);
    char buf[100];
    Vec2 wsm = WindowFromWorldCoords(world_mouse);
    sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f   wsm: %5.1f, %5.1f",
            world_mouse.x(),world_mouse.y(),screen_mouse.x(),screen_mouse.y(),wsm.x(),wsm.y());
    DrawString(50,50,buf);
*/

/*
	Thing *playerThing = thingGetByName( "Player*" );
	if( playerThing && playerThing->player )
	{
		KenDrawString( 90, 4, f( "HEALTH: %d", (int)(playerThing->player->health*100.0) ) );
	}

	Thing *scoreThing = thingGetByName( "Score*" );
	if( scoreThing && scoreThing->score )
	{
		KenDrawString( 90, 6, f( "SCORE: %d", (int)(scoreThing->score->score*100.0) ) );
	}
*/

	KenDrawString( 50-10, 4, f( "LEVEL: %d", gameFlow->level ) );
	KenDrawString( 50-10, 8, f( "GAS: %d", (int)gameFlow->bonus ) );
	KenDrawString( 80, 4, f( "SCORE: %d", gameFlow->score ) );
//	KenDrawString( 10, 4, f( "EYES SHUT: %d", gameFlow->eyeCount - gameFlow->eyeCountClear ) );

	if( !gameFlow->gameRunning )
	{
		gameFlow->draw();
	}

	if( gameFlow->levelUp )
	{
		gameFlow->levelUp->draw();
	}


    //DrawTimingBars();
    SwapBuffers(WindowDC);

	gameFlow->endProcessing();

	if( !gameFlow->gameRunning )
	{
		for (int i=0; i<PS2Controllers.num_joysticks(); i++)
		{
			if (PS2Controllers.button_went_down(Joystick::BUTTON_START, i))
			{
				RestartGame();
				break;
			}
		}
	}

}

/*
void forceAdd( char *nameWild, float angleDegrees, float forceMagnitude )
{
	Thing *thing = thingGetByName( nameWild );
	assert( thing );
	if( thing )
	{
		Handle thingHandle( *thing->rb, Vec2(0.0f,0.0f) );
		Control_Force cf( thingHandle, Globals.GuyMoveMaxVelocity, forceMagnitude );
		
		float angleRads = degsToRads( angleDegrees );
		cf.m_target = Vec2( cos( angleRads ), sin( angleRads ) );
		
		int dummy=1;
	}	
}
*/

void mouseJump( char *nameWild, Vec2 mouseDown, Vec2 mouseScreen )
{
	Thing *jumper = thingGetByName( nameWild );
	assert( jumper );

	Vec2 screen;
	screen.m_v[0] = (float)WindowWidth;
	screen.m_v[1] = (float)WindowHeight;
	
	Vec2 mouse;
	mouse.m_v[0] = mouseScreen.m_v[0] - mouseDown.m_v[0];
	mouse.m_v[1] = - ( mouseScreen.m_v[1] - mouseDown.m_v[1] );
	
	float xPct = mouse.m_v[0] / screen.m_v[0];
	float yPct = mouse.m_v[1] / screen.m_v[1];

	Vec2 percentVector( xPct, yPct );
	float jumpMag = min( 1.0f, percentVector.len() / Globals.GuyJumpScreenPercent );

	Vec2 jumpVector = percentVector.norm();
	jumpVector *= ( jumpMag * Globals.GuyJumpScalar );

	float jumpAngleRadians = atan2( jumpVector.m_v[1], jumpVector.m_v[0] );
	float jumpAngleDegrees = radiansToDegrees( jumpAngleRadians );

	jumper->jumpAngle     = jumpAngleDegrees;
	jumper->jumpMagnitude = jumpVector.len();
}

int mouseGrabPlayer( HWND hwnd )
{
	if( mouseThing )
		return 0;

	SetCapture(hwnd);

    Vec2 bodySpacePosition;
    Thing *player = rigidBodyToThing( simulator.pick( bodySpacePosition, mouseWorld ) );
    if( player )
    {
		if( !player->isName( "player*" ) )
			return 1;
    
		mouseThing = player;
        mouseSpring = new Spring_Force(
            Handle(worldBody, mouseWorld),
            Handle(*mouseThing->rb, bodySpacePosition), 
            Globals.MouseSpringKs,
            Globals.MouseSpringKd,
            Globals.MouseSpringRestLength
            );
        forceList.push_back( mouseSpring );
        simulator.add( *mouseSpring );
    }
	return 1;
}

void mouseGrabRelease( WPARAM wParam )
{
	if(!(wParam & MK_MBUTTON & MK_RBUTTON))
	{
		ReleaseCapture();
	}

    if( mouseSpring )
    {
		forceList.erase( std::find( forceList.begin(), forceList.end(), mouseSpring ) );
        simulator.remove(*mouseSpring);
        delete mouseSpring;
        mouseSpring = 0;
        
        mouseThing = 0;
    }

}



/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	static float plowLeftBase = 0.0f;
	static float plowRightBase = 0.0f;

    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
/*
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            mouseWorld = WorldFromWindowCoords(mouseScreen);
*/
            break;
        }

        case WM_LBUTTONDOWN:
        {
			if( !mouseGrabPlayer( hwnd ) )
			{
				Thing *tongue = thingGetByName( "tongue*" );
				if( tongue )
					tongue->tryUnStick();
			}
        } break;

        case WM_LBUTTONUP:
        {
			//mouseGrabRelease( wParam );
        } break;

        case WM_RBUTTONDOWN:
        {
			mouseGrabRelease( wParam );
        } break;
        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;

        case WM_KEYUP: {
			switch( wParam ) {
				case 'a': case 'A': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
                    player->setPlow( "plow_left", plowLeftBase, 1.0 );
					break;
				}
				case 'd': case 'D': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
                    player->setPlow( "plow_right", plowRightBase, 1.0 );
					break;
				}
				case VK_LEFT: case VK_RIGHT: {
					Thing *player = thingGetByName( "player*" );
					if( player )
						player->setTurn( 0.0 );
					break;
				}
				case VK_UP: case VK_DOWN: {
					Thing *player = thingGetByName( "player*" );
					if( player )
					{
						player->setPush( 0.0 );
						soundStop( player->player->backupbeep );
						player->player->engine.play( "long_1" );
					}
					break;
				}
			}
		} break;

            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_LEFT: {
                    //CameraAngleRadians += float((5.0/180.0)*M_PI);
					Thing *player = thingGetByName( "player*" );
                    if( player )
						player->setTurn( Globals.PlowTurnRate );
                    break;
                }
                case VK_RIGHT: {
                    //CameraAngleRadians -= float((5.0/180.0)*M_PI);
					Thing *player = thingGetByName( "player*" );
                    if( player )
						player->setTurn( -Globals.PlowTurnRate );
                    break;
                }

                case VK_UP: {
                    //CameraAngleRadians -= float((5.0/180.0)*M_PI);
					Thing *player = thingGetByName( "player*" );
                    if( player )
                    {
						player->setPush( Globals.PlowPushRate );
						player->player->engine.play( "long_2" );
					}
                    break;
                }
                case VK_DOWN: {
                    //CameraAngleRadians -= float((5.0/180.0)*M_PI);
					Thing *player = thingGetByName( "player*" );
                    if( player )
                    {
						player->setPush( -Globals.PlowPushRate );
                        soundPlayUnique( player->player->backupbeep, "backupbeep" );
					}
                    break;
                }

				case ' ': {
					if( !gameFlow->gameRunning )
					{
						RestartGame();
					}
					break;
				}
				

                case 'A': case 'a': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
                    plowLeftBase = player->setPlow( "plow_left", plowLeftBase, Globals.PlowTug );
                    break;
                }

                case 'D': case 'd': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
                    plowRightBase = player->setPlow( "plow_right", plowRightBase, Globals.PlowTug );
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

                case 'E': {
					mouseGrabRelease( 0 );
                    launchEditor(TestLevelFileName);
                    ClearAll();
                    LoadTestFile();
                } break;
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
//        case WM_KEYUP:
            break;

            // mouse
        case WM_MBUTTONDOWN:
        case WM_MBUTTONUP:
            break;

        case WM_MOUSEMOVE: {
            mouseScreen = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
            mouseWorld = WorldFromWindowCoords(mouseScreen);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 screen )
{
    float xn = float(screen.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-screen.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
