/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_control_force.h"

#include "StrExtra.h"

using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

/********************************* globals **********************************/

#define PI 3.1415926f


// window/camera stuff
char const AppName[] = "Hot n Cold Super Flea";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 100.0;
Vec2 WorldFromWindowCoords( Vec2 screen );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(.05f, .05f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);

char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;
	
	float GuyMoveMaxVelocity;
	float GuyMoveSpringConstant;

	float GuyJumpScalar;
	float GuyJumpScreenPercent;
};

static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

float degreesToRadians( float degs )
{
	float rads = PI/180.0f*degs;
	return rads;
}

float radiansToDegrees( float rads )
{
	float degs = 180.0f/PI*rads;
	if( degs < 0.0f )
		degs += 360.0f;
	if( degs > 360.0f )
		degs -= 360.0f;
	return degs;
}

float angleAdd( float angleDegrees, float angleAddDegrees )
{
	float degs = angleDegrees + angleAddDegrees;
	if( degs < 0.0f )
		degs += 360.0f;
	if( degs > 360.0f )
		degs -= 360.0f;
	return degs;
}


class Thing;
Thing *thingGetContactAtAngle( Thing *asker, float angleDegrees, float angleToleranceDegrees );


//-----------------------------------------------
// Thing
//
class Thing
{
public:
    Rigid_Body *rb;

	Control_Force *walkForce;
	int walkDir;
	float jumpAngle;
	float jumpMagnitude;
    
	Thing( char *texture_file , Rigid_Body *_rb )
	{
		walkForce = 0;
		walkDir = 0;
		jumpAngle = 0.0f;
		jumpMagnitude = 0.0f;
		rb = _rb;
		assert(rb);
		rb->m_pAppData = this;
		simulator.add(*rb);
	}

	int isName( char *nameWild )
	{
		assert( rb );
		return strMatchWild( nameWild, (char*)rb->name().c_str() );
	}

	void walk( int _dir )
	{
		walkDir = _dir;
	}

	void jumpApply( float angleDegrees, float magnitude )
	{
		float angleOppositeDegrees = angleAdd( angleDegrees, 180.0f );
		Thing *floor = thingGetContactAtAngle( this, angleOppositeDegrees, 180.0f );
		if( floor )
		{
			float angleRadians = degreesToRadians( angleDegrees );
			Vec2 jumpVector( cos( angleRadians ), sin( angleRadians ) );
			jumpVector *= magnitude;
			rb->v = jumpVector;
		}
	}

	
	void walkApply()
	{
		if( walkDir == 0 )
		{
			if( walkForce )
			{
				simulator.remove( *walkForce );
				delete walkForce;
				walkForce = 0;
			}
			return;
		}

		float myFastestVelocity = Globals.GuyMoveMaxVelocity;

		if( !walkForce )
		{
			Handle thingHandle( *rb, Vec2(0.0f,0.0f) );
			walkForce = new Control_Force( thingHandle, myFastestVelocity, Globals.GuyMoveSpringConstant );
			simulator.add( *walkForce );
		}

		float friction = 1.0f;
		// If I'm not standing on anything, then I can't 'step on it' to propel 
		// myself forward anymore.  So, if my velocity vector is upward, no friction
		// is allowed.
//		if( rb->v.y() > 0.0 )
//		{
//			friction = 0.0f;
//		}
//		else
		{
			// Find the thing that is directly below me.
			Thing *floor = thingGetContactAtAngle( this, 270.0f, 45.0f );
			if( !floor )
			{
				friction = 0.0;
			}
			else
			{
				friction = floor->rb->k_friction;
			}
		}		
		
		walkForce->m_max_v = myFastestVelocity * friction;

		float angleDegrees = (walkDir == 1) ? 0.0f : 180.0f;
		float angleRads = degreesToRadians( angleDegrees );
		walkForce->m_target = Vec2( cos( angleRads ), sin( angleRads ) );
	}

	void skidApply()
	{
		Thing *floor = thingGetContactAtAngle( this, 270.0f, 45.0f );
		if( floor )
		{
			float friction = floor->rb->k_friction;
			rb->v.m_v[0] *= friction;
		}
		
//		float angleOppositeDegrees = angleAdd( angleDegrees, 180.0f );
//		Thing *floor = thingGetContactAtAngle( this, angleOppositeDegrees, 180.0f );
//		if( floor )
	}
	
	void process()
	{
		skidApply();
		walkApply();
		if( jumpMagnitude > 0.0f )
		{
			jumpApply( jumpAngle, jumpMagnitude );
			jumpMagnitude = 0.0f;
		}
	}

    ~Thing()
    {
		delete rb;
		rb = 0;
    }
};

typedef vector<Thing*> ThingList;
ThingList thingList;

ThingList thingGetByNameWild( char *nameWild )
{
	ThingList list;
	for( int i=0 ; i<thingList.size() ; ++i )
	{
		Thing *thing = thingList[i];
		assert( thing );
		if( thing->isName( nameWild ) )
		{
			list.push_back( thing );
		}
	}
	return list;
}

Thing *thingGetByName( char *nameWild )
{
	for( int i=0 ; i<thingList.size() ; ++i )
	{
		Thing *thing = thingList[i];
		assert( thing );
		if( thing->isName( nameWild ) )
		{
			return thing;
		}
	}
	return 0;
}

Thing *thingGetContactAtAngle( Thing *asker, float angleDegrees, float angleToleranceDegrees )
{
	for( Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); ++cit )
	{
        Rigid_Body *r1 = (Rigid_Body *)(cit->m_Prb_a);
        Rigid_Body *r2 = (Rigid_Body *)(cit->m_Prb_b);
        if( ( r1 == asker->rb || r2 == asker->rb ) && r1 && r2 )
        {
			Vec2 contactPointInBodySpace = asker->rb->world_space_to_body_space( cit->m_x );
			float contactAngleRadians = atan2( contactPointInBodySpace.m_v[1], contactPointInBodySpace.m_v[0] );
			float contactAngleDegrees = radiansToDegrees( contactAngleRadians );
			float delta = fabs( angleDegrees - contactAngleDegrees );
			if( delta < angleToleranceDegrees || delta > 360 - angleToleranceDegrees )
			{
				if( asker->rb == r1 )
					return (Thing *)r2->m_pAppData;
				else
					return (Thing *)r1->m_pAppData;
			}
		}
	}
	return 0;
}

//-----------------------------------------------
// Forces
//
typedef vector<Binary_Force*> ForceList;
ForceList forceList;

static Vec2 mouseDown,mouseScreen,mouseWorld;

static Binary_Force *spring = 0;


GLuint
LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    verify(LoadBitmapByExtension(Filename,&result));

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

/*
void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}
*/

void
ClearAll(void)
{
//    ClearMouseSpring();
    
    for(int i = 0;i < thingList.size();++i)
    {
        assert(thingList[i]);
        simulator.remove(*thingList[i]->rb);
    }

    for(int i = 0;i < forceList.size();++i)
    {
        assert(forceList[i]);
        simulator.remove(*forceList[i]);
    }

    thingList.clear();
    forceList.clear();
}

void
LoadTestFile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(TestLevelFileName, 0, 0, 1.0f/10.0f, 1.0f/16.0f,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

    for (int i=0; i < num_rb; ++i)
    {
        thingList.push_back(new Thing("content/bitmaps/test.jpg",rb_list[i]));
    }

    for (int i=0; i < num_force; ++i)
    {
        forceList.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }
}

/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
    Globals.Gravity = -40.0f;

    Globals.MouseSpringKs = 4.0f;
    Globals.MouseSpringKd = 2.0f;
    Globals.MouseSpringRestLength = 0.0f;

    Globals.GuyMoveMaxVelocity    = 40.0f;
    Globals.GuyMoveSpringConstant = 30.0f;

	Globals.GuyJumpScreenPercent  = 0.50f;	// Mouse over 20% of the screen for a 100% jump.
	Globals.GuyJumpScalar         = fabs(Globals.Gravity) * 4.0f;	// Your boost at a 100% jump.
    
    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "Segment1.sgp", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);

    LoadTestFile();
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

	for( int i=0 ; i<thingList.size() ; i++ )
	{
		thingList[i]->process();
	}

    gravity.m_g.set_y(Globals.Gravity);
    
    /**************************** step the simulator ****************************/   

//    if (spring)                             //update
//        spring->m_a = Handle(world_body, world_mouse);

    // @todo
    simulator.step(0.01f);

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();

    // draw mouse
/*
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();
*/

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // iterate all the forces, rendering them
    glColor3f(1,0,0);
    for(int i = 0;i < forceList.size();++i)
    {
        assert(forceList[i]);
        Binary_Force *Force = forceList[i];

        glForce1v(Force);
    }
    glForce1v(spring); 
	
	// iterate all the bodies, rendering them
    glColor3f(0,0,0);
    for(int i = 0;i < thingList.size();++i) {
        assert(thingList[i]);
        Thing &thing = *thingList[i];

        glBody1v(thing.rb->m_Pbody.get(), 1);
    }

    
    glDisable(GL_BLEND);

/*
    glColor3f(0, 0, 0);
    char buf[100];
    Vec2 wsm = WindowFromWorldCoords(world_mouse);
    sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f   wsm: %5.1f, %5.1f",
            world_mouse.x(),world_mouse.y(),screen_mouse.x(),screen_mouse.y(),wsm.x(),wsm.y());
    DrawString(50,50,buf);
*/

    DrawTimingBars();
    SwapBuffers(WindowDC);
}

/*
void forceAdd( char *nameWild, float angleDegrees, float forceMagnitude )
{
	Thing *thing = thingGetByName( nameWild );
	assert( thing );
	if( thing )
	{
		Handle thingHandle( *thing->rb, Vec2(0.0f,0.0f) );
		Control_Force cf( thingHandle, Globals.GuyMoveMaxVelocity, forceMagnitude );
		
		float angleRads = degsToRads( angleDegrees );
		cf.m_target = Vec2( cos( angleRads ), sin( angleRads ) );
		
		int dummy=1;
	}	
}
*/

void walk( char *nameWild, int dir )
{
	Thing *thing = thingGetByName( nameWild );
	assert( thing );

	thing->walk( dir );
}

void mouseJump( char *nameWild, Vec2 mouseDown, Vec2 mouseScreen )
{
	Thing *jumper = thingGetByName( nameWild );
	assert( jumper );

	Vec2 screen;
	screen.m_v[0] = (float)WindowWidth;
	screen.m_v[1] = (float)WindowHeight;
	
	Vec2 mouse;
	mouse.m_v[0] = mouseScreen.m_v[0] - mouseDown.m_v[0];
	mouse.m_v[1] = - ( mouseScreen.m_v[1] - mouseDown.m_v[1] );
	
	float xPct = mouse.m_v[0] / screen.m_v[0];
	float yPct = mouse.m_v[1] / screen.m_v[1];

	Vec2 percentVector( xPct, yPct );
	float jumpMag = min( 1.0f, percentVector.len() / Globals.GuyJumpScreenPercent );

	Vec2 jumpVector = percentVector.norm();
	jumpVector *= ( jumpMag * Globals.GuyJumpScalar );

	float jumpAngleRadians = atan2( jumpVector.m_v[1], jumpVector.m_v[0] );
	float jumpAngleDegrees = radiansToDegrees( jumpAngleRadians );

	jumper->jumpAngle     = jumpAngleDegrees;
	jumper->jumpMagnitude = jumpVector.len();
}



/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_LBUTTONDOWN:
        {
			mouseDown = mouseScreen;
			SetCapture(hwnd);
        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

            mouseJump( "Player", mouseDown, mouseScreen );
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

			if(spring)
            {
                forceList.push_back(spring);
                spring = 0;
            }
        } break;
        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;

        case WM_KEYUP: {
			switch( wParam ) {
				case 'a': case 'A': case 'd': case 'D': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
					player->walk( 0 );
					break;
				}
			}
		} break;

            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'A': case 'a': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
					player->walk( -1 );
                    break;
                }

                case 'D': case 'd': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
					player->walk( 1 );
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

                case 'E': {
                    launchEditor(TestLevelFileName);
                    ClearAll();
                    LoadTestFile();
                } break;
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
//        case WM_KEYUP:
            break;

            // mouse
        case WM_MBUTTONDOWN:
        case WM_MBUTTONUP:
            break;

        case WM_MOUSEMOVE: {
            mouseScreen = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
            mouseWorld = WorldFromWindowCoords(mouseScreen);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 screen )
{
    float xn = float(screen.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-screen.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
