/*
 * Indie Game Jame 2 - Ken's "Broken Dodgers"
 * 2004/3/15 checker@d6.com
 * 2004/3/16 ken@demarest.com
 */
 
/*

Other thoughts:
- defend a critter as it crosses the screen
- walk in the shadow of larger, slow moving things
- touch power ups to get health, 
- some things are not harmfull, can be pushed to then move other dangerous stuff
	- some (or all) of these are fixed
- generators shooting out evil
	- lines
	- clouds
	- duration on life of evil
	- basically, any particle generator
*/ 
 
 
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_control_force.h"

#include "StrExtra.h"
#include "SysParse.h"
#include "FastRandom.h"
#include "SysMalloc.h"

using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

#include "tempstr.h"

#include "../../utils/misc/framerate.h"

/********************************* globals **********************************/

#define PI 3.1415926f

// window/camera stuff
char const AppName[] = "Broken Dodgers";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 100.0;
Vec2 WorldFromWindowCoords( Vec2 screen );
Vec2 WindowFromWorldCoords( Vec2 world );

GLuint LoadTexture(char *Filename);

float frameTime = 0.0f;
float frameTimeDelta = 1.0f;

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(.05f, .05f);    //a little drag helps with stability
//Gravity_Force gravity(Vec2(0.0f, -1.0f));
Gravity_Force gravity(Vec2(0.0f, 0.0f));
Rigid_Body    worldBody(make_Body_Null(),Vec2(0,0),0,0,0);

char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct GlobalPropertyStruct
{
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;
	
	float GuyMoveMaxVelocity;
	float GuyMoveSpringConstant;

	float GuyJumpScalar;
	float GuyJumpScreenPercent;

	float PowerUpHealth;

	float OutOfWorld;
};

static GlobalPropertyStruct Globals;
static property *globalProperties;
static property_system_editor *GlobalsEditor;

float degreesToRadians( float degs )
{
	float rads = PI/180.0f*degs;
	return rads;
}

float radiansToDegrees( float rads )
{
	float degs = 180.0f/PI*rads;
	if( degs < 0.0f )
		degs += 360.0f;
	if( degs > 360.0f )
		degs -= 360.0f;
	return degs;
}

float angleAdd( float angleDegrees, float angleAddDegrees )
{
	float degs = angleDegrees + angleAddDegrees;
	if( degs < 0.0f )
		degs += 360.0f;
	if( degs > 360.0f )
		degs -= 360.0f;
	return degs;
}


//-----------------------------------------------
// Forces
//
typedef vector<Binary_Force*> ForceList;
ForceList forceList;


class Thing;
typedef vector<Thing*> ThingList;

class Hit;
typedef vector<Hit*> HitList;

Thing * thingGetContactAtAngle( Thing *asker, float angleDegrees, float angleToleranceDegrees );
HitList thingGetHitList( Thing *asker );
Thing * thingCreate( char *fileName, Vec2 position, Vec2 velocity );
Binary_Force *forceAdd( Thing *thing1, Thing *thing2, float Ks, float Kd, float restLength = 1.0f );
void forceRemove( Binary_Force *force );

Thing *rigidBodyToThing( Rigid_Body *rb )
{
	if( !rb )
		return 0;

	return (Thing *)rb->m_pAppData;
}

class Hit
{
public:
	Thing *thing;
	Contact *contact;
	float angleDegrees;

	Hit( Thing *_thing, Contact *_contact )
	{
		thing = _thing;
		contact  = _contact;
	}
};


class Player
{
public:
	float health;
	
	Player()
	{
		health = 1.0;
	}

	void healthAdjust( float healing )
	{
		health += healing;
		
		if( health < 0.0 )
		{
			health = 0.0f;
		}

		if( health > 1.0  )
		{
			health = 1.0f;
		}
	}

	void harm( float damage )
	{
		healthAdjust( -damage );
	}
	
};

void creaturePick( char *name )
{
	switch( fastRandomInt(4) )
	{
		case 0:
			strcpy( name, "beetle" );
			break;
		case 1:
			strcpy( name, "spider" );
			break;
		case 2:
			strcpy( name, "fly" );
			break;
		case 3:
			strcpy( name, "ladybug" );
			break;
	}
}

class Generator
{
public:
	Thing *parent;
	double nextGeneration;
	double period;

	Generator( Thing *_parent, float _period )
	{
		assert( _parent );
		parent = _parent;
		period = _period;
		nextGeneration = frameTime + 0.0f; //period;
	}
	
	int readyToGenerate()
	{
		return frameTime > nextGeneration;
	}
	
	void generate( char *s, Vec2 position, Vec2 velocity )
	{
		char creatureName[1024];
		strcpy( creatureName, s );
		if( !stricmp( creatureName, "any" ) )
		{
			creaturePick( creatureName );	
		}
	
		thingCreate( creatureName, position, velocity );
	}

	void testGeneration( char *myName, Vec2 position, Vec2 velocity )
	{
		if( readyToGenerate() )
		{
			char temp[1024];
			char *s = sysParse( 1024, temp, myName, "_" );
			generate( s, position, velocity );
			nextGeneration = frameTime + period;
		}
	}
};


class Tongue
{
public:
	Thing *parent;
	Thing *thingAttached;
	Binary_Force *forceAttaching;
	double timeLastUnstuck;
	float suctionStrength;
	
	Tongue( Thing *_parent )
	{
		assert( _parent );
		parent = _parent;
		thingAttached = 0;
		forceAttaching = 0;
		timeLastUnstuck = 0.0f;
		suctionStrength = 0.0f;
	}
	int isStuck()
	{
		return thingAttached != 0;
	}
	int isSticky()
	{
		if( frameTime < timeLastUnstuck + 1.0 )
			return 0;
		return 1;
	}
	void stickTo( Thing *_thing, Hit *_hit, float _strength )
	{
		assert( thingAttached == 0 );
		assert( forceAttaching == 0 );
		assert( frameTime >= timeLastUnstuck + 1.0f );
		
		suctionStrength = _strength;
		
		thingAttached = _thing;
		
		forceAttaching = forceAdd( parent, thingAttached, 100.0f, 20.0f, 0.0f );
		assert( forceAttaching );
	}

	void unStick()
	{
		assert( forceAttaching );
		forceRemove( forceAttaching );
		delete forceAttaching;
		forceAttaching = 0;
		thingAttached = 0;
		suctionStrength = 0.0f;
		
		timeLastUnstuck = frameTime;
	}

	int shouldLoosen( float externalForces )
	{
		if( externalForces > suctionStrength )
		{
			return 1;
		}
		return 0;
	}
};

/*
Creature types
	Fly - flies quickly, eratically, keeps flying if you don't hit it solidly
		  enough with your tongue to knock it out.  Gets away in short time unless
		  you've stunned it.
	Beetle - slow, heavy, never gets away, but its weight can cause your suction
		  to not be strong enough.
	Spider - medium sized, medium mover, if it touches your head then it stings
		  you and harms you. When it stings you let go.
	Ladybug - benign, medium speed eratic mover, but very small.


*/
PROP_STRUCT struct CreaturePropertyStruct
{
	float points;
	float speed;
	float leanSwitch;
	float turnRateMax;
};

float roll( float occurEveryXSeconds )
{
	return( fastRandomFloat( occurEveryXSeconds ) < frameTimeDelta );
}


class Creature
{
public:
	Thing *parent;
	CreaturePropertyStruct props;

	float lean;

	Creature( Thing *_parent, char *creatureName )
	{
		assert( _parent );
		parent = _parent;
		
		property *templateGroup = FindSubPropertyByName( LocalTemplates, "CreaturePropertyStruct" );
		property *masterGroup   = FindSubPropertyByName( globalProperties, "CreatureData" );
		property *dataGroup     = FindSubPropertyByName( masterGroup, creatureName );
		ConvertPropertyToStruct( templateGroup, dataGroup, &props );

		lean = 1.0f;
	}

	float getSpeed()
	{
		return props.speed;
	}
	
	float getPoints()
	{
		return props.points;
	}

	void move( Vec2 &velocity, float &heading )
	{
		if( props.speed > 0.0f )
		{
			if( roll(props.leanSwitch) )
				lean = -lean;
				
			if( velocity.len() < props.speed )
			{
				velocity.m_v[0] += cos( heading );
				velocity.m_v[1] += sin( heading );
			}
			float turnAmount = PI/2.0f / (props.turnRateMax / frameTimeDelta);
			heading += turnAmount * lean;
		}
	}
};

class Score
{
public:
	float score;
	Score()
	{
		score = 0.0f;
	}
	void scoreAdjust( float points )
	{
		score += points;
	}
};

//-----------------------------------------------
// Thing
//
class Thing
{
public:
    Rigid_Body *rb;
	Player *    player;
	Generator * generator;
	Tongue *    tongue;
	Creature *  creature;
	Score *     score;

	GLuint texture;

	float jumpAngle;
	float jumpMagnitude;

	float health;

	int alive;
    
	Thing( Rigid_Body *_rb )
	{
		alive = 1;
		player = 0;
		generator = 0;
		tongue = 0;
		creature = 0;
		score = 0;

		texture = 0;

		jumpAngle = 0.0f;
		jumpMagnitude = 0.0f;
		rb = _rb;
		assert(rb);
		rb->m_pAppData = this;

		texture = LoadTexture( getName() );

		if( isName( "player*" ) )
			player = new Player;

		if( isName( "gen_*" ) )
			generator = new Generator( this, 10.0 );

		if( isName( "tongue*" ) )
			tongue = new Tongue( this );

		if( isName( "score*" ) )
			score = new Score();

		if( isCreature() )
			creature = new Creature( this, getName() );
		
		simulator.add(*rb);
	}

    ~Thing()
    {
		simulator.remove(*rb);
    
		if( player ) delete player;
		if( generator ) delete generator;
		if( tongue ) delete tongue;
		if( creature ) delete creature;
		if( score ) delete score;
		delete rb;
		rb = 0;
    }

	int isDead()
	{
		return !alive;
	}

	int isName( char *nameWild )
	{
		assert( rb );
		return strMatchWild( nameWild, (char*)rb->name().c_str() );
	}

	char *getName()
	{
		static char temp[1024];
		assert( rb );
		strcpy( temp, (char*)rb->name().c_str() );
		return &(temp[0]);
	}

	void jumpApply( float angleDegrees, float magnitude )
	{
		float angleOppositeDegrees = angleAdd( angleDegrees, 180.0f );
		Thing *floor = thingGetContactAtAngle( this, angleOppositeDegrees, 180.0f );
		if( floor )
		{
			float angleRadians = degreesToRadians( angleDegrees );
			Vec2 jumpVector( cos( angleRadians ), sin( angleRadians ) );
			jumpVector *= magnitude;
			rb->v = jumpVector;
		}
	}

	void analyzeHits()
	{
		assert( player );
		HitList hitList = thingGetHitList( this );
		for( int i=0 ; i<hitList.size() ; ++i )
		{
			Hit *hit = hitList[i];
				
			if( hit->thing->isName( "evil*" ) )
			{
				player->harm( 1.0f );
			}
			if( hit->thing->isName( "health*" ) )
			{
				player->healthAdjust( Globals.PowerUpHealth );
				hit->thing->die();
			}
		}
	}

	void tryUnStick()
	{
		if( tongue && tongue->isStuck() )
		{
			assert( tongue );
			tongue->unStick();
			assert( tongue->thingAttached == 0 );
		}
	}

	int isCreature()
	{
		return isName( "gem*" ) || isName( "spider*" ) || isName( "beetle*" ) || isName( "fly*" ) || isName( "ladybug*" );
	}

	void trySticking()
	{
		assert( tongue );
		if( tongue->isStuck() || !tongue->isSticky() )
			return;
			
		HitList hitList = thingGetHitList( this );
		for( int i=0 ; i<hitList.size() ; ++i )
		{
			Hit *hit = hitList[i];
				
			if( hit->thing->isCreature() )
			{
				Vec2 vecSum = rb->x + hit->thing->rb->x;
			
				tongue->stickTo( hit->thing, hit, vecSum.len() );
				return;
			}
		}
	}

	void scorePoints()
	{
		assert( score );
		
		HitList hitList = thingGetHitList( this );
		for( int i=0 ; i<hitList.size() ; ++i )
		{
			Hit *hit = hitList[i];
			if( hit->thing->isCreature() && !hit->thing->isDead() )
			{
				score->scoreAdjust( hit->thing->creature->getPoints() );
				hit->thing->die();
				
			}
		}			
			
	}


	void die()
	{
		alive = 0;
	}

	void checkOutOfWorld()
	{
		Vec2 dist = rb->x;
		if( dist.len() > Globals.OutOfWorld )
			die();
	}

	
	void process()
	{
		if( isDead() )
			return;
			
		if( jumpMagnitude > 0.0f )
		{
			jumpApply( jumpAngle, jumpMagnitude );
			jumpMagnitude = 0.0f;
		}

		if( player )
			analyzeHits();
		
		if( generator )
			generator->testGeneration( getName(), rb->x, rb->v );
		
		if( tongue )
		{
			trySticking();

			if( tongue->isStuck() )
			{
				Vec2 vecSum = tongue->thingAttached->rb->x;
				if( tongue->shouldLoosen( vecSum.len() ) )
					tryUnStick();
			}
		}

		if( creature )
		{
			creature->move( rb->v, rb->O.m_z );
		}

		if( score )
			scorePoints();

		checkOutOfWorld();
	}

};

ThingList thingList;

ThingList thingGetByNameWild( char *nameWild )
{
	ThingList list;
	for( int i=0 ; i<thingList.size() ; ++i )
	{
		Thing *thing = thingList[i];
		assert( thing );
		if( thing->isName( nameWild ) )
		{
			list.push_back( thing );
		}
	}
	return list;
}

Thing *thingGetByName( char *nameWild )
{
	for( int i=0 ; i<thingList.size() ; ++i )
	{
		Thing *thing = thingList[i];
		assert( thing );
		if( thing->isName( nameWild ) )
		{
			return thing;
		}
	}
	return 0;
}

Thing *thingGetContactAtAngle( Thing *asker, float angleDegrees, float angleToleranceDegrees )
{
	for( Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); ++cit )
	{
        Rigid_Body *r1 = (Rigid_Body *)(cit->m_Prb_a);
        Rigid_Body *r2 = (Rigid_Body *)(cit->m_Prb_b);
        if( ( r1 == asker->rb || r2 == asker->rb ) && r1 && r2 )
        {
			Vec2 contactPointInBodySpace = asker->rb->world_space_to_body_space( cit->m_x );
			float contactAngleRadians = atan2( contactPointInBodySpace.m_v[1], contactPointInBodySpace.m_v[0] );
			float contactAngleDegrees = radiansToDegrees( contactAngleRadians );
			float delta = fabs( angleDegrees - contactAngleDegrees );
			if( delta < angleToleranceDegrees || delta > 360 - angleToleranceDegrees )
			{
				if( asker->rb == r1 )
					return (Thing *)r2->m_pAppData;
				else
					return (Thing *)r1->m_pAppData;
			}
		}
	}
	return 0;
}

//
// Warning: does not return 'dead' things.
//
HitList thingGetHitList( Thing *asker )
{
	HitList hitList;
	
	for( Contact_List::const_iterator cit=simulator.m_contacts.begin(); cit != simulator.m_contacts.end(); ++cit )
	{
        Rigid_Body *r1 = (Rigid_Body *)(cit->m_Prb_a);
        Rigid_Body *r2 = (Rigid_Body *)(cit->m_Prb_b);
        if( ( r1 == asker->rb || r2 == asker->rb ) && r1 && r2 )
        {
			Thing *thingHit = ( asker->rb == r1 ) ? rigidBodyToThing(r2) : rigidBodyToThing(r1);
			if( !thingHit->isDead() )
			{
				hitList.push_back( new Hit( thingHit, cit._Myptr ) ); //cit ) );
			}
		}
	}
	return hitList;
}


static Vec2 mouseDown,mouseScreen,mouseWorld;
static Thing *mouseThing = 0;
static Binary_Force *mouseSpring = 0;

struct TextureCache
{
	char name[80];
	int  texture;
};	

typedef vector<TextureCache*> TextureCacheList;
TextureCacheList textureCacheList;

int textureCacheGetByName( char *name )
{
	for( int i=0 ; i<textureCacheList.size() ; ++i )
	{
		if( !stricmp( textureCacheList[i]->name, name ) )
			return textureCacheList[i]->texture;
	}
	return 0;
}

GLuint LoadTexture(char *fileName)
{
    GLuint texture = textureCacheGetByName( fileName );
    if( texture )
		return texture;
    
	f path( "content\\bitmaps\\%s.psd", fileName );
    
    Bitmap_Load_Result result;
    int ok = LoadBitmapByExtension( path.s, &result );
    
    if( !ok )
		return 0;

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

	TextureCache *textureCache = new TextureCache;
	textureCache->texture = texture;
	strcpy( textureCache->name, fileName );
	textureCacheList.push_back( textureCache );

    return textureCache->texture;
}

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

Binary_Force *forceAdd( Thing *thing1, Thing *thing2, float Ks, float Kd, float restLength )
{

    Binary_Force *f = new Spring_Force(
        Handle( *thing1->rb, Vec2(0.0f,0.0f) ),
        Handle( *thing2->rb, Vec2(0.0f,0.0f) ), 
        Ks,
        Kd,
        restLength
        );
    forceList.push_back( f );
    simulator.add( *f );
	return f;
}

void forceRemove( Binary_Force *force )
{
	ForceList::iterator bfi = std::find( forceList.begin(), forceList.end(), force );
	forceList.erase( bfi );
	simulator.remove( *force );
}


/*
void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}
*/

void
ClearAll(void)
{
//    ClearMouseSpring();
    
    for(int i = 0;i < thingList.size();++i)
    {
        assert(thingList[i]);
        simulator.remove(*thingList[i]->rb);
    }

    for(int i = 0;i < forceList.size();++i)
    {
        assert(forceList[i]);
        simulator.remove(*forceList[i]);
    }

    thingList.clear();
    forceList.clear();
}

Thing *thingCreate( char *fileName, Vec2 position, Vec2 velocity )
{
	Thing *thingCreated = 0;

    int num_rb=0, num_force=0;
    Rigid_Body **rb_list = 0;
    Binary_Force **force_list = 0;

	f path( "content\\critters\\%s.physdb", fileName );

    loadPhysicsFile(path.s, 0, 0, 1.0f/10.0f, 1.0f/16.0f,
                    &rb_list, &num_rb, &force_list, &num_force, &worldBody);

    for (int i=0; i < num_rb; ++i)
    {
		rb_list[i]->x = position;
		rb_list[i]->v = velocity;
		
        Thing *addMe = new Thing( rb_list[i] );
		addMe->rb->x = position;
		addMe->rb->v = velocity;
        if( i == 0 )
        {
			thingCreated = addMe;
		}
        thingList.push_back( addMe );
    }

    for (int i=0; i < num_force; ++i)
    {
        forceList.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }
    
	return thingCreated;
}

void
LoadTestFile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(TestLevelFileName, 0, 0, 1.0f/10.0f, 1.0f/16.0f,
                    &rb_list, &num_rb, &force_list, &num_force, &worldBody);

    for (int i=0; i < num_rb; ++i)
    {
		Thing *made = new Thing( rb_list[i] );
		thingList.push_back( made );
    }

    for (int i=0; i < num_force; ++i)
    {
        forceList.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }
}


/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
	sysMallocInit( 0, 0, 0, 0, 0, 0 );


	fastRandomize();
    InitializeFrameRate(30.0f);

    Globals.Gravity = 0.0f; //-40.0f;

    Globals.MouseSpringKs = 100.0f;
    Globals.MouseSpringKd = 20.0f;
    Globals.MouseSpringRestLength = 0.0f;

    Globals.GuyMoveMaxVelocity    = 40.0f;
    Globals.GuyMoveSpringConstant = 30.0f;

	Globals.GuyJumpScreenPercent  = 0.50f;	// Mouse over 20% of the screen for a 100% jump.
	Globals.GuyJumpScalar         = fabs(Globals.Gravity) * 4.0f;	// Your boost at a 100% jump.

	Globals.PowerUpHealth         = 0.2f;
	Globals.OutOfWorld			  = 200.0f;
    
    globalProperties = ReadPropertyFromFile( "default.igj2props" );

    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, globalProperties);
    
    frameTime = GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "Segment1.sgp", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);


    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(worldBody);

    LoadTestFile();
}

void KenDrawString( int xPct, int yPct, char *s )
{
	float x = (float)(xPct)*(float)WindowWidth/100.0f;
	float y = (float)(100-yPct)*(float)WindowHeight/100.0f;
	glColor3f( 0.0f, 0.0f, 0.0f );
	
	DrawString( x, y, s );
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
	
    float newFrameTime = GetTime(); // throw away initial 0.0 so we're seconds-from-startup
    frameTimeDelta = newFrameTime - frameTime;
    frameTime = newFrameTime;

    ConvertPropertyToStruct(
		// This is the struct in this file to find, by struct type.
		FindSubPropertyByName(LocalTemplates, "GlobalPropertyStruct"),
		// This is the actual properties inside the master list
        FindSubPropertyByName(globalProperties,"GlobalPropertyData"),
        &Globals);


	ThingList killList;
	for( int i=0 ; i<thingList.size() ; i++ )
	{
		thingList[i]->process();
		if( thingList[i]->isDead() )
			killList.push_back( thingList[i] );

		Thing *thingSource = thingList[i];
		if( thingSource->isName( "make_*" ) )
		{
			char name[1024];
			strcpy( name, thingSource->getName() );
			char dummy[1024];
			char *nameToMake = sysParse( 1024, dummy, name, "_" );
			thingCreate( nameToMake, thingSource->rb->x, thingSource->rb->v );
			killList.push_back( thingSource );
		}
	}

	for( int i=0 ; i<killList.size() ; i++ )
	{
		Thing *thingKill = killList[i];
		delete thingKill;
		thingList.erase( std::find( thingList.begin(), thingList.end(), thingKill ) );
	}


    gravity.m_g.set_y(Globals.Gravity);
    
    /**************************** step the simulator ****************************/   

    if( mouseSpring )                             //update
        mouseSpring->m_a = Handle(worldBody, mouseWorld);

    int const NumberOfSimSteps = 1;
    for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
        // step as many times as you like
        simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
        // do any game stuff you want here (like per-step collision response, etc.)
    }
    // now wait for the rest of the time so we're consistent on fast machines
    WaitFrameRate(GetFrameRate());

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();

    // draw mouse
/*
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();
*/

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // iterate all the forces, rendering them
    glColor3f(1,0,0);
    for(int i = 0;i < forceList.size();++i)
    {
        assert(forceList[i]);
        Binary_Force *Force = forceList[i];

        glForce1v(Force);
    }
//    glForce1v(mouseSpring); 
	
	// iterate all the bodies, rendering them
    glColor3f(0,0,0);
    for(int i = 0;i < thingList.size();++i) {
        assert(thingList[i]);
        Thing &thing = *thingList[i];

        glBody1v(thing.rb->m_Pbody.get(), 1);
    }

    
    glDisable(GL_BLEND);

/*
    glColor3f(0, 0, 0);
    char buf[100];
    Vec2 wsm = WindowFromWorldCoords(world_mouse);
    sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f   wsm: %5.1f, %5.1f",
            world_mouse.x(),world_mouse.y(),screen_mouse.x(),screen_mouse.y(),wsm.x(),wsm.y());
    DrawString(50,50,buf);
*/

	Thing *playerThing = thingGetByName( "Player*" );
	if( playerThing && playerThing->player )
	{
		KenDrawString( 90, 4, f( "HEALTH: %d", (int)(playerThing->player->health*100.0) ) );
	}

	Thing *scoreThing = thingGetByName( "Score*" );
	if( scoreThing && scoreThing->score )
	{
		KenDrawString( 90, 6, f( "SCORE: %d", (int)(scoreThing->score->score*100.0) ) );
	}

    DrawTimingBars();
    SwapBuffers(WindowDC);
}

/*
void forceAdd( char *nameWild, float angleDegrees, float forceMagnitude )
{
	Thing *thing = thingGetByName( nameWild );
	assert( thing );
	if( thing )
	{
		Handle thingHandle( *thing->rb, Vec2(0.0f,0.0f) );
		Control_Force cf( thingHandle, Globals.GuyMoveMaxVelocity, forceMagnitude );
		
		float angleRads = degsToRads( angleDegrees );
		cf.m_target = Vec2( cos( angleRads ), sin( angleRads ) );
		
		int dummy=1;
	}	
}
*/

void mouseJump( char *nameWild, Vec2 mouseDown, Vec2 mouseScreen )
{
	Thing *jumper = thingGetByName( nameWild );
	assert( jumper );

	Vec2 screen;
	screen.m_v[0] = (float)WindowWidth;
	screen.m_v[1] = (float)WindowHeight;
	
	Vec2 mouse;
	mouse.m_v[0] = mouseScreen.m_v[0] - mouseDown.m_v[0];
	mouse.m_v[1] = - ( mouseScreen.m_v[1] - mouseDown.m_v[1] );
	
	float xPct = mouse.m_v[0] / screen.m_v[0];
	float yPct = mouse.m_v[1] / screen.m_v[1];

	Vec2 percentVector( xPct, yPct );
	float jumpMag = min( 1.0f, percentVector.len() / Globals.GuyJumpScreenPercent );

	Vec2 jumpVector = percentVector.norm();
	jumpVector *= ( jumpMag * Globals.GuyJumpScalar );

	float jumpAngleRadians = atan2( jumpVector.m_v[1], jumpVector.m_v[0] );
	float jumpAngleDegrees = radiansToDegrees( jumpAngleRadians );

	jumper->jumpAngle     = jumpAngleDegrees;
	jumper->jumpMagnitude = jumpVector.len();
}

int mouseGrabPlayer( HWND hwnd )
{
	if( mouseThing )
		return 0;

	SetCapture(hwnd);

    Vec2 bodySpacePosition;
    Thing *player = rigidBodyToThing( simulator.pick( bodySpacePosition, mouseWorld ) );
    if( player )
    {
		if( !player->isName( "player*" ) )
			return 1;
    
		mouseThing = player;
        mouseSpring = new Spring_Force(
            Handle(worldBody, mouseWorld),
            Handle(*mouseThing->rb, bodySpacePosition), 
            Globals.MouseSpringKs,
            Globals.MouseSpringKd,
            Globals.MouseSpringRestLength
            );
        forceList.push_back( mouseSpring );
        simulator.add( *mouseSpring );
    }
	return 1;
}

void mouseGrabRelease( WPARAM wParam )
{
	if(!(wParam & MK_MBUTTON & MK_RBUTTON))
	{
		ReleaseCapture();
	}

    if( mouseSpring )
    {
		forceList.erase( std::find( forceList.begin(), forceList.end(), mouseSpring ) );
        simulator.remove(*mouseSpring);
        delete mouseSpring;
        mouseSpring = 0;
        
        mouseThing = 0;
    }

}



/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            mouseWorld = WorldFromWindowCoords(mouseScreen);
            break;
        }

        case WM_LBUTTONDOWN:
        {
			if( !mouseGrabPlayer( hwnd ) )
			{
				Thing *tongue = thingGetByName( "tongue*" );
				if( tongue )
					tongue->tryUnStick();
			}
        } break;

        case WM_LBUTTONUP:
        {
			//mouseGrabRelease( wParam );
        } break;

        case WM_RBUTTONDOWN:
        {
			mouseGrabRelease( wParam );
        } break;
        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;

        case WM_KEYUP: {
			switch( wParam ) {
				case 'a': case 'A': case 'd': case 'D': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
					break;
				}
			}
		} break;

            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'A': case 'a': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
                    break;
                }

                case 'D': case 'd': {
                    Thing *player = thingGetByName( "player*" );
                    assert( player );
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

                case 'E': {
					mouseGrabRelease( 0 );
                    launchEditor(TestLevelFileName);
                    ClearAll();
                    LoadTestFile();
                } break;
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
//        case WM_KEYUP:
            break;

            // mouse
        case WM_MBUTTONDOWN:
        case WM_MBUTTONUP:
            break;

        case WM_MOUSEMOVE: {
            mouseScreen = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
            mouseWorld = WorldFromWindowCoords(mouseScreen);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 screen )
{
    float xn = float(screen.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-screen.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
