#include "parseany.h"
#include "string.h"
#include "stdlib.h"
#include "SysAssert.h"
#include "SysParse.h"
#include "minmax.h"

ParseWord::ParseWord( CharPtr &src, char *separatorList, int allowSingleQuote )
{
	sysAssert( strlen(src) < PARSEWORD_LEN );
	src = sysParse( PARSEWORD_LEN, string, src, separatorList, allowSingleQuote );
}

ParseInt::ParseInt( CharPtr &src, char *separatorList, int allowSingleQuote )
{
	ParseWord word( src, separatorList, allowSingleQuote );
	number = atoi( word.string );
}

ParseList::ParseList( CharPtr &src, int index, char *separatorList, int allowSingleQuote )
{
	string[0] = '\0';
	while( index >= 0 )
	{
		ParseWord word( src, separatorList, allowSingleQuote );
		--index;
		if( index < 0 )
		{
			sysAssert( strlen(word) < PARSELIST_LEN );
			strcpy( string, word );
		}
	}
}

ParseLine::ParseLine( CharPtr &src, int pixelLength, int appendCR, char *separatorList )
{
	while( *src && (*src == ' ' || *src == '\r') )
		++src;
	int len = strlen( src );
	int offset = 0;
	int copyLen = min( pixelLength, len - offset );
	while( strchr(separatorList, src[offset+copyLen])==0 && src[offset+copyLen] != '\0' )
		--copyLen;
	char *where = strchr( src, '\n' );
	if( where && where < (src+copyLen) )
		copyLen = (where - src) + 1;

	strncpy( string, src+offset, copyLen );
	int apLen = copyLen;
	if( appendCR )
		if( apLen > 0 && string[apLen-1] != '\n' )
			string[apLen++] = '\n';
	string[apLen] = '\0';
	src += copyLen;
}

FieldParser::FieldParser( char *_fieldString, char _quote )
{
	fieldString = _fieldString;
	quote = _quote;
}

char *FieldParser::get( char *fieldName )
{
	buffer[0] = '\0';

	char field[1024];
	char value[1024];
	char *s = fieldString;

	while( s && *s )
	{
		s = sysParse( 1024, field, s, "=", 1, 1, quote );
		s = sysParse( 1024, value, s, " ", 1, 1, quote );

		if( !stricmp( field, fieldName ) )
		{
			strcpy( buffer, value );
			break;
		}
	}
	return buffer;
}

