#ifndef PARSEANY_H
#define PARSEANY_H

//
// These routines will parse a word out of a phrase.  If enclosed in quotes
// this will figure it out and give you the quoted object.  Single quotes
// are allowed if you enable them with the allowSingleQuote parameter.
//
// The 'src' must be a char * that you're willing to have advanced past the
// word and separator.
//
// If you don't pass a separator in, the standard separators are " \t\n,"
//
// If you need to do something funky, call parseInto() directly.  Otherwise
// you can make a call like:
//
// char *commandString = "copy \"jolly roger.jpg\" c:\newrog.jpg";
// char *s = commandString;
// ParseWord cmd( s, " " );			// cmd.string = copy
// ParseWord srcFile( s, " " );		// cmd.string = jolly roger.jpg
// ParseWord destFile( s, " " );	// cmd.string = c:\newrog.jpg
//

typedef char *CharPtr;

struct ParseWord
{
	#define PARSEWORD_LEN	(1024*16)
	char string[PARSEWORD_LEN];
	ParseWord( CharPtr &src, char *separatorList = 0, int allowSingleQuote = 0 );
	operator char * () { return string; }
};

struct ParseInt
{
	int number;
	ParseInt( CharPtr &src, char *separatorList = 0, int allowSingleQuote = 0 );
	operator int () { return number; }
};

struct ParseList
{
	#define PARSELIST_LEN	(1024*16)
	char string[PARSELIST_LEN];
	ParseList( CharPtr &src, int index, char *separatorList = 0, int allowSingleQuote = 0 );

	char *getString()	{ return string; }
	ParseInt getInt()	{ char *s = string; return ParseInt(s); }
};


struct ParseLine
{
	char string[1024*16];
	ParseLine( CharPtr &src, int pixelLength, int appendCR, char *separatorList );
};

class FieldParser
{
	char *fieldString;
	char buffer[1024];
	char quote;

public:
	FieldParser( char *_fieldString, char _quote = 0 );
	char *get( char *fieldName );
};



#endif
