#include "ctype.h"
#include "stdlib.h"

#include "StaticString.h"
#include "SysAssert.h"
#include "Elipsis.h"
#include "SysMalloc.h"

// Static strings have two characters between them, so that
// we can do verification to check for over-runs.

#define VERIFY_CHAR		(-43)

int   staticStringMaxLen = 0;
char *staticStringGlobal = 0;
char *curPtr = 0;
char *priorPtr = 0;

struct StaticStringDeconstructor
{
	~StaticStringDeconstructor()
	{
		if( staticStringGlobal )
		{
			delete staticStringGlobal;
			staticStringGlobal = 0;
		}
	}
};
StaticStringDeconstructor staticStringDeconstructor;



void staticStringInit( int maxLen )
{
	if( staticStringGlobal )
	{
		delete staticStringGlobal;
		staticStringGlobal = 0;
	}
	staticStringMaxLen = maxLen;
	staticStringGlobal = new char[staticStringMaxLen];
	memset( staticStringGlobal, 0, staticStringMaxLen );

	curPtr = staticStringGlobal;
	*curPtr = VERIFY_CHAR;
	++curPtr;
}


char *staticString( int allocLen )
{
	if( !staticStringGlobal )
		staticStringInit( 1024*16 );

	sysAssert( *(curPtr-1) == VERIFY_CHAR );
		// If this happens your prior staticString suffered an over-run.

	sysAssert( allocLen < staticStringMaxLen >> 2 );
		// It is just plain bad karma to allocate more than one fourth
		// the the entire static string area. You're too likely to
		// run right back over it.

	if( curPtr + allocLen + 1 >= staticStringGlobal + staticStringMaxLen ) {
		curPtr = staticStringGlobal;
		*curPtr = VERIFY_CHAR;
		++curPtr;
	}
	priorPtr = curPtr;
	curPtr += allocLen;
	*(curPtr++) = VERIFY_CHAR;

	return priorPtr;
}

char *staticString( char *s )
{
	if( !s )
		return 0;
	char *result = staticString( strlen(s)+1 );
	strcpy( result, s );
	return result;
}

char *staticStringAssemble( int allocLen, char *fmt, ... )
{
	char *result = staticString( allocLen );
	elipsisTo( (unsigned int)allocLen, result, fmt );
	return result;
}

