#include "StrExtra.h"
#include "ctype.h"
#include "SysAssert.h"
#include "string.h"

void strReplace( char *s, char a, char b ) {
	while( s && *s )
	{
		if( *s == a )
			*s = b;
		++s;
	}
}

void strReplacePairs( char *s, char *c )
{
	if( !c )
		return;

	while( c[0] && c[1] )
	{
		strReplace( s, c[0], c[1] );
		c += 2;
	}
}

char *strSkipWhite( char *src )
{
	while( *src && ( *src == ' ' || *src == '\t' || (*src < 0) ) )
	{
		++src;
	}
	return src;
}

char *strTrim( char *src, char *trimList )
{
	char *start = src;
	while( *src ) ++src;

	if( src > start ) --src;

	while( src > start && ( strchr( trimList, *src ) || (*src < 0) ) )
	{
		*src = '\0';
		--src;
	}
	return src;
}

char *strTrimWhite( char *src )
{
	char *start = src;
	while( *src ) ++src;

	if( src > start ) --src;

	while( src > start && ( *src == ' ' || *src == '\t' || (*src < 0) ) )
	{
		*src = '\0';
		--src;
	}
	return src;
}


char *stristr( char *src, char *findMe )
{
	if( !src )
	{
		return 0;
	}

	sysAssert( findMe );

	char *s = src;
	char firstLetter = tolower(*findMe);

	while( *s )
	{
		if( tolower(*s) == firstLetter )
		{
			char *a = s;
			char *b = findMe;
			while( *b && *a && tolower(*a) == tolower(*b) )
			{
				++a;
				++b;
			}
			if( !(*b) )
			{
				return s;
			}
		}
		++s;
	}
	return 0;
}

char *strzcpy( char *dest, char *src, int len )
{
	sysAssert( len >= 0 );
	memcpy( dest, src, len );
	dest[len-1] = '\0';
	return dest;
}

int strzcmp( char *main, char *subStr )
{
	return strnicmp( main, subStr, strlen(subStr) );
}


char *strcpyUntil( char *dest, char *src, char *stoppers )
{
	while( *src && !strchr( stoppers, *src ) )
	{
		*dest = *src;
		++src;
		++dest;
	}
	*dest = '\0';

	return src;
}

char *strzcpyUntil( char *dest, char *src, int maxLen, char *stoppers )
{
	while( *src && --maxLen && !strchr( stoppers, *src ) )
	{
		*dest = *src;
		++src;
		++dest;
	}
	*dest = '\0';

	return src;
}

char *strzcpyUntilChar( char *dest, char *src, int maxLen, char stopper )
{
	while( *src && --maxLen && *src != stopper )
	{
		*dest = *src;
		++src;
		++dest;
	}
	*dest = '\0';
	return src;
}

char *strcpyCap( char *dest, char *src )
{
	char *result = dest;
	while( *src )
	{
		*dest = toupper( *src );
		++src;
		++dest;
	}
	*dest = '\0';
	return result;
}


char *strGet( char *list, int index, char indexSeparator )
{
	static char result[256];

	char *temp = new char[strlen(list)+1];
	strcpy( temp, list );
	char *s = temp;
	char *found = 0;
	while( s && index >= 0 )
	{
		if( index == 0 )
		{
			found = s;
		}
		s = strchr( s, indexSeparator );
		if( s && *s == indexSeparator )
		{
			*s='\0'; 
			++s;
		}
		--index;
	}

	result[0] = '\0';
	if( found )
	{
		strcpy( result, found );
	}
	delete temp;
	return result;
}


int strCount( char *buffer, char indexSeparator )
{
	int count = 0;
	char *s = buffer;
	if( *s )
	{
		++count;
	}
	while( *s )
	{
		if( *s == indexSeparator )
		{
			++count;
		}
		++s;
	}

	return count;
}

int strFind ( char *what, char *list, char indexSeparator )
{
	int count = strCount( list, indexSeparator );
	for( int i=0 ; i<count ; ++i )
	{
		if( !stricmp( what, strGet( list, i, indexSeparator ) ) )
			return i;
	}
	return -1;
}


char *strAdvanceUntil( char *s, char *findOneOf )
{
	while( *s && !strIs( *s, findOneOf ) )
	{
		++s;
	}
	return s;
}

char *strSkip( char *src, char *skipUs )
{
	if( src )
	{
		while( *src && ( strchr( skipUs, *src ) || (*src<0) ) )
		{
			++src;
		}
	}
	return src;
}

int strMatchWild( char *srcWild, char *other )
{
	sysAssert( srcWild );
	if( !other )
		return 0;
	char *a = other;
	char *b = srcWild;
	char *bLastStar = 0;
	char *aLastStopper = 0;

	while( *b )
	{
		if( *b == '*' )
		{
			bLastStar = b;
			++b;
			char stopper = tolower(*b);
			while( *a && tolower(*a) != stopper )
			{
				++a;
			}
			if( tolower(*a) != stopper )
			{
				return 0;
			}
			aLastStopper = a;
		}
		else
		{
			if( tolower(*b) != tolower(*a) )
			{
				if( bLastStar )
				{
					b = bLastStar;
					a = aLastStopper + 1;
					continue;
				}
				return 0;
			}
			++a;
			++b;
		}
	}
	return (*a) == '\0';
}

char *strRetreatUntil( char *s, char *start, char *findMe )
{
	if( s > start && !(*s) )
		--s;
	while( s != start && strchr(findMe,*s)==0 )
	{
		--s;
	}
	return s;
}

