//
// Str.cpp - Str class source
// (c) 1997 Ken Demarest
//

//
// Include files
//

#include "string.h"
#include "Elipsis.h"
#include "TempStr.h"
#include "ctype.h"
#include "stdlib.h"
#include "StrExtra.h"
#include "SysMalloc.h"
#include "minmax.h"

//
// Routines
//

#define MAXLEN  (1024*32)

char *Str::empty = "";

int isHex( char c )
{
	return (c>='0' && c<='9') || (c>='a' && c<='f') || (c>='A' && c<='F');
}



f::f( char *fmt, ... )
{
	char temp[MAXLEN];

	// Looks like this look finds instances that look like %4F or %2D and then
	// confirms that the character after the % is within legal hex range for
	// being less than 127.
	for( int i=0 ; i<(int)strlen(fmt) ; ++i )
	{
		if( fmt[i] == '%' )
		{
			if( isHex(fmt[i+1]) && isHex(fmt[i+2]) )
			{
				// I think I need this better explained to me (by me!) -Ken
				if( fmt[i+1] < '0' || fmt[i+1] > '9' )
					sysAssert(0);
			}
		}
	}

	elipsisTo(MAXLEN,temp,fmt);

	int len = strlen(temp)+1;
	s = (char*)sysMalloc( max(1024,len) );
	memcpy( s, temp, len );
}

f::~f()
{
	if( s && s!=empty )
		sysFree( s );
}

void Str::mapTo( char *_s )
{
	if( s && s!=empty )
		sysFree( s );
	s = _s;
}

void Str::_set( char *newValue, int length )
{
	if( s && s!=empty )
		sysFree( s );

	if( !newValue )
		s = 0;
	else
	{
		s = (char*)sysMalloc( max(1024,length+1) );
		memcpy( s, newValue, length );
		s[length] = '\0';
	}
}

void Str::_set( char *newValue )
{
	if( s && s!=empty )
		sysFree( s );

	if( !newValue )
		s = 0;
	else
	{
		int len = strlen(newValue)+1;
		s = (char*)sysMalloc( max(1024,len) );
		memcpy( s, newValue, len );
	}
}

Str::Str( int _length )
{
	s = (char*)sysMalloc( _length );
	if( _length )
		s[0] = '\0';
}

Str::Str( f &_f )
{
	if( s && s!=empty )
		sysFree( s );
	s = _f.take();	
}


Str::~Str()
{
	if( s && s!=empty )
		sysFree( s );
}

void Str::append( char *addMe )
{
	if( addMe && *addMe )
	{
		int myLen  = ( s ? strlen(s) : 0 );
		int addLen = strlen( addMe );

		char *temp = new char[ myLen+addLen+1 ];

		if( s )
			memcpy( temp, s, myLen );

		memcpy( temp+myLen, addMe, addLen+1 );	// +1 gets the '\0' at the end.

		mapTo( temp );
	}
}


int Str::is( char *other )
{
	if( !other )
		return s == other;
	if( !s )
		return 0;
	return !stricmp(s,other);
}

int Str::isWild( char *other )
{
	if( !other )
		return s == other;
	if( !s )
		return 0;
	return strMatchWild( other, s );
}

int Str::getLength()
{
	return s ? strlen(s) : 0;
}

char *Str::allCaps()
{
	if( s )
	{
		char *ss = s;
		while( *ss )
		{
			*ss = toupper(*ss);
			++ss;
		}
	}
	return s;
}

char *Str::allLower()
{
	if( s )
	{
		char *ss = s;
		while( *ss )
		{
			*ss = tolower(*ss);
			++ss;
		}
	}
	return s;
}

char *Str::cap()
{
	if( s )
	{
		char *ss = s;
		while( *ss && !isalpha(*ss) )
		{
			++ss;
		}

		if( *ss )
		{
			*ss = toupper(*ss);
		}
	}
	return s;
}

int Str::getInt()
{
	if( !s )
		return 0;
	return atoi( s );
}



/*
void Strf::set( char *fmt, ... )
{
	char temp[MAXLEN];
	elipsisTo(MAXLEN,temp,fmt);
	_set( temp );
}

void Strf::append( char *fmt, ... )
{
	char temp[MAXLEN];
	strcpy( temp, s );
	int length = strlen(temp);
	int limit = MAXLEN-(length+1);
	char *ptr = temp + length;
	#pragma warning(disable:4018)		// warning C4018: '<' : signed/unsigned mismatch
	elipsisTo(limit,ptr,fmt);
	#pragma warning(default:4018)
	_set( temp );
}
*/

