#include "malloc.h"

#include "stdio.h"
#include "string.h"
#include "assert.h"

#include "SysAssert.h"
#include "Lock.h"
#include "LeakTracker.h"
#include "Log.h"
#include "FileReport.h"

#include "WinPrintVc.h"

//#define warn(exp) assert(exp)
//(void)( (exp) || (geAssertEntryPoint(#exp, __FILE__, __LINE__), 0) )

int memLock = 0;

int newCount = 0;
int newBytes = 0;
int delCount = 0;
int delBytes = 0;

int inMalloc = 0;

int curTouchNum = 1;

int doKenHeapChecking = 1;

#define MEM_TELL_FILE 1

#if MEM_TELL_FILE
	#define CALLER_FILE_LEN (40)
	char *_km = 0;
#endif

char *memLog = "+memLog";



#define PREFIX_SIZE            (4)

// This is because VC's debug allocation routines ALSO put their
// own filler characters in as a prefix - four bytes worth, in
// fact, so we've got to fill 4 less when !debug.
#ifndef DEBUG
	#define PREFIX_FILL_SIZE       (PREFIX_SIZE)
#else
	#define PREFIX_FILL_SIZE       (PREFIX_SIZE-4)
#endif

#define POSTFIX_SIZE           (8)

// Here are the stolen definitions from VC's C runtime lib.
//static unsigned char _bNoMansLandFill = 0xFD;   /* fill no-man's land with this */
//static unsigned char _bDeadLandFill   = 0xDD;   /* fill free objects with this */
//static unsigned char _bCleanLandFill  = 0xCD;   /* fill new objects with this */


#pragma pack(1)
struct KenHeader
{
	unsigned char fillArea[PREFIX_FILL_SIZE];	
	KenHeader *next;
	KenHeader *prev;
	#ifdef MEM_TELL_FILE
		char callerFile[CALLER_FILE_LEN];
	#endif
	int caller;
	int size;
	int touchNum;
	unsigned char *getDataStart()
	{
		return ((unsigned char *)this) + sizeof(*this);
	}
};
#pragma pack()


struct MemEvent
{
	KenHeader *ptr;
	char *name;
	void set( KenHeader *_ptr, char *_name ) { ptr=_ptr; name=_name; }
};

static int recentAllocCount = 0;
static MemEvent recentAlloc[16] = {0,};
static int recentFreeCount = 0;
static MemEvent recentFree[16] = {0,};

static KenHeader *head = NULL;

/*
void kHeapTouch()
{
	LOCK(memLock);
	KenHeader *cur = head;
	while( cur ) {
		if( cur->touchNum != 9999 ) {
			cur->touchNum++;
		}
		cur = cur->next;
	}
}
*/


int kHeapCountTouch( int touchNum )
{
	LOCK(memLock);
	KenHeader *cur = head;
	int count = 0;
	while( cur )
	{
		if( cur->touchNum == touchNum )
		{
			++count;
		}
		cur = cur->next;
	}
	return count;
}

//extern "C" void winPrintVc( char *fmt, ... );

void kHeapDump( char *destFile, int showOnlyTouchNum )
{
	{
		LOCK(memLock);
		FILE *f = 0;
		if( destFile )
		{
			f = fopen( destFile, "w+t" );
			if( fileReportOpen )
				fileReportOpen( (int)f, destFile, 0, 1, 1 );
		}

		KenHeader *cur = head;
		int count = 0;
		if( f ) fprintf( f, "BEGIN\n" );
		else    winPrintVc( "BEGIN\n" );

		if( f ) fprintf( f, " --Addr--  -Caller-   --Size--   (#)\n" );
		else    winPrintVc( " --Addr--  -Caller-   --Size--   (#)\n" );

		while( cur )
		{
			if( (showOnlyTouchNum==0) || (cur->touchNum==showOnlyTouchNum) )
			{
				char data[1024];
				char *s = (char*)cur->getDataStart();
				char *dest = data;
				char *hex = "01234567890ABCDEF";
				for( int i=0 ; i<(cur->size<15?cur->size:15) ; ++i )
				{
					if( s[i] >= ' ' && s[i] <= '~' )
					{
						*dest++ = s[i];
					}
					else
					{
						unsigned char c = s[i];
						*dest++ = hex[c/16];
						*dest++ = hex[c%16];
					}
				}
				*dest = '\0';

				if( f ) 
					fprintf( f, "%010x %08X   %8X   (%d) %s\n", cur->getDataStart(), cur->caller, cur->size, cur->touchNum, data );
				else    
					winPrintVc( "%010x %08X   %8X   (%d) %s\n", cur->getDataStart(), cur->caller, cur->size, cur->touchNum, data );
			}
			cur = cur->next;
			count++;
		}
		if( f )
		{
			fprintf( f, "END\n" );
			if( fileReportClose )
				fileReportClose( (int)f, destFile );
			fclose( f );
		}
		else
		{
			winPrintVc( "END\n" );
		}
	}
}

void kHeapDump()
{
	kHeapDump( 0, 0 );
}

int kHeapContains( void *memblock )
{
	LOCK(memLock);
	KenHeader *cur    = head;
	KenHeader *found  = 0;
	KenHeader *findMe = (KenHeader *)(((char*)memblock) - sizeof(KenHeader));

	while( cur )
	{
		if( cur == findMe )
		{
			sysAssert( !found );
			found = cur;
		}
		cur = cur->next;
	}
	sysAssert( found );
	return( found != 0 );
}

void validateHeapEntry( KenHeader *ptr )
{
	unsigned char *prefix = (unsigned char *)ptr;
	prefix -= ( PREFIX_SIZE - PREFIX_FILL_SIZE );

	try {
		for( int n=0 ; n<PREFIX_SIZE ; ++n )
		{
			if( *prefix != 0xfd )
			{
				logAdd( memLog, "Corrupt heap prefix %p\n", ptr );
				sysAssert( 0 );
			}
			++prefix;
		}
	}
	catch(...)
	{
		__asm int 3;
	}


	unsigned char *postfix = ptr->getDataStart() + ptr->size;
	try
	{
		for( int n=0 ; n<POSTFIX_SIZE ; ++n )
		{
			if( *postfix != 0x1c )
			{
				logAdd( memLog, "Corrupt heap post %p\n", ptr );
				sysAssert( 0 );
			}
			++postfix;
		}
	}
	catch(...)
	{
		__asm int 3;
	}

	if( ptr->next )
	{
		if( ptr->next->prev != ptr )
		{
			logAdd( memLog, "Corrupt heap next link %p\n", ptr );
		}
		sysAssert( ptr->next->prev == ptr );
	}

	if( ptr->prev )
	{
		if( ptr->prev->next != ptr )
		{
			logAdd( memLog, "Corrupt heap prev link %p\n", ptr );
		}
		sysAssert( ptr->prev->next == ptr );
	}
}


int kHeapChk()
{
	if( !doKenHeapChecking )
	{
		return 1;
	}

	LOCK(memLock);
	KenHeader *cur = head;
	int count = 0;
	while( cur )
	{
		validateHeapEntry( cur );
		cur = cur->next;
		++count;
	}

	return( count == newCount - delCount );
}



void *kMallocInternal( unsigned int cb )
{
//	sysAssert( cb > 0 );
// malloc is allowed to allocate a zero-sized object. 

	if( cb == 0x800 )
	{
		cb = 0x800;
	}

	cb += sizeof(KenHeader) + POSTFIX_SIZE;

	volatile int *currBP;
	__asm mov [currBP], ebp;
	int caller;
	caller = *(currBP+1);

	if( inMalloc )
	{
		unsigned char *ptr = (unsigned char *)currBP;
		// 12 is the magic number when you're in release.
		#ifndef _DEBUG
			caller = *((int*)(ptr+12));
		#else
			caller = *((int*)(ptr+28));
		#endif
	}

	unsigned char *q = (unsigned char *)::malloc( cb ); //return _nh_malloc( cb, 1 );
	try
	{
		sysAssert( q );
	}
	catch(...)
	{
		__asm int 3;
	}
	memset( q, 0x1c, cb );

	cb -= ( sizeof(KenHeader) + POSTFIX_SIZE );

	KenHeader *me = (KenHeader *)q;
	me->caller = caller;
	#ifdef MEM_TELL_FILE 
		strncpy( me->callerFile, _km ? _km : "?", CALLER_FILE_LEN-1 );
		me->callerFile[CALLER_FILE_LEN-1] = '\0';
	#endif
	me->next = NULL;
	me->prev = NULL;
	me->size = cb;
	me->touchNum = curTouchNum; //(inMalloc?100:0);
	memset( me->fillArea, 0xfd, PREFIX_FILL_SIZE );
	if( head ) {
		head->prev = me;
		me->next = head;
	}
	head = me;

	newBytes += me->size;
	newCount += 1;

	inMalloc = 0;

	recentAllocCount = (recentAllocCount+1)&0x0F;
	recentAlloc[recentAllocCount].set( me, _km );

	return q + sizeof(KenHeader);
}

void kFreeInternal( void * p )
{
	KenHeader *me = (KenHeader *)( (char*)p - sizeof(KenHeader) );

	// Slows down the world!
	// validateHeapEntry( me );

	// on free
	if( me->prev )
	{
		me->prev->next = me->next;
	}
	if( me->next )
	{
		me->next->prev = me->prev;
	}
	if( me == head )
	{
		head = me->next;
	}

	delCount++;
	delBytes += me->size;

	sysAssert( delCount <= newCount );

	recentFreeCount = (recentFreeCount+1)&0x0F;
	recentFree[recentFreeCount].set( me, _km );

	unsigned char *freeMe = (unsigned char *)me;
	::free( freeMe );
}

void kMemSafe()
{
	head->touchNum = 9999;
}


void *kMalloc( size_t size )
{
	LOCK(memLock);
	inMalloc = 1;
	return kMallocInternal( size );
}

void *kCalloc( size_t num, size_t size )
{
	LOCK(memLock);
	inMalloc = 1;
	char *ptr = (char*)kMallocInternal( num * size );
	if( ptr )
	{
		memset( ptr, 0, num * size );
	}
	return ptr;
}

void *kMallocZeroAllowed( size_t size )
{
	LOCK(memLock);
	if( size  == 0 )
	{
		return NULL;
	}
	return kMalloc( size  );
}


void kFree( void *memBlock )
{
	sysAssert( kHeapChk() );

	// This will slow everything down!
	// sysAssert( kHeapContains( memBlock ) );

	LOCK(memLock);
	kFreeInternal( memBlock );
}

void *kRealloc( void *memblock, size_t size )
{

	KenHeader *me = 0;
	
	if( memblock )
	{
		me = (KenHeader *)( (char *)memblock - sizeof(KenHeader) );
	}

	char *newMemblock = (char *)kMalloc( size );
	if( newMemblock && me )
	{
		LOCK(memLock);
		unsigned int copyAmount = ( (signed int)size < me->size ) ? size : me->size;
		memcpy( newMemblock, (char *)memblock, copyAmount );
		kFreeInternal( memblock );
	}

	return newMemblock;
}

int kGetAllocCount()
{
	return newCount - delCount;
}

int kGetAllocBytes()
{
	return newBytes - delBytes;
}

void *hackMalloc( char *fileLine, int bytes )
{
	extern char *_km;
	_km = fileLine;
	return kMalloc( bytes );
}
