

#include "SysMalloc.h"
#include "string.h"		// for sysStrDup.
#include "malloc.h"
#include "SysAssert.h"
#include "cExterns.h"

int sysMallocInitialized = 0;

// These are the pointers to the actual calls.
SysCallocFn   *sysCallocCall   = 0;
SysMallocFn   *sysMallocCall   = 0;
SysReallocFn  *sysReallocCall  = 0;
SysFreeFn     *sysFreeCall     = 0;
SysHeapChkFn  *sysHeapChkCall  = 0;
SysHeapDumpFn *sysHeapDumpCall = 0;

// This fn is used to initialize the memory management routines.
 
void sysMallocInit( 
		SysCallocFn   *calloc, 
		SysMallocFn   *malloc, 
		SysReallocFn  *realloc,
		SysFreeFn     *free,
		SysHeapChkFn  *heapChk,
		SysHeapDumpFn *heapDump
	)
{
	if( calloc )   sysCallocCall = calloc;
	if( malloc )   sysMallocCall = malloc;
	if( realloc )  sysReallocCall = realloc;
	if( free )     sysFreeCall = free;
	if( heapChk )  sysHeapChkCall = heapChk;
	if( heapDump ) sysHeapDumpCall = heapDump;

	ccCalloc = calloc;
	ccMalloc = malloc;
	ccRealloc = realloc;
	ccFree = free;
	ccHeapChk = heapChk;
	ccHeapDump = heapDump;

	sysMallocInitialized = 1;
}

#ifndef BOUNDS_CHECKER

void * sysCalloc    ( size_t num, size_t size )
{
	sysAssert( sysMallocInitialized );
	if( !sysCallocCall )	return calloc( num, size );
	return sysCallocCall( num, size );
}

#ifdef sysMalloc
	#undef sysMalloc
#endif
 
void * sysMalloc    ( size_t size )
{
	sysAssert( sysMallocInitialized );
	return new char[size];
}

void * sysMalloc2( char *file, size_t size )
{
//	extern char *km;
//	km = file;
	void *result = sysMalloc( size );
//	km = 0;
	return result;
}
 
void * sysRealloc   ( void *memblock, size_t size )
{
	sysAssert( sysMallocInitialized );
	if( !sysReallocCall ) return realloc( memblock, size );
	return sysReallocCall( memblock, size );
}

 
void   sysFree      ( void *memblock )
{																	
	// sysAssert( sysMallocInitialized );
	delete memblock;
}

 
int    sysHeapChk   ( void )
{
	sysAssert( sysMallocInitialized );
	if( !sysHeapChkCall ) { return _heapchk() == _HEAPOK; } 
	return sysHeapChkCall();
}

 
void   sysHeapDump  ( void )
{
	sysAssert( sysMallocInitialized );
	if( !sysHeapDumpCall ) return;
	sysHeapDumpCall();
}


 
char *sysStrDup( const char *src )
{
	int len = 0;
	char *dest = 0;

	if( !src )
	{
		return 0;
	}
	len = strlen(src) + 1;
	dest = new char[len];
	memcpy( dest, src, len );
	return dest;
}




#endif
