#ifndef SYSMALLOC_H
#define SysMALLOC_H

//#define BOUNDS_CHECKER 1

/*
This module intends to overload common memory allocation
methods found in all C runtime libraries.
*/

// This is straight from VC's stdlib.h
#ifndef _SIZE_T_DEFINED
	typedef unsigned int size_t;
	#define _SIZE_T_DEFINED
#endif

// TYPE DEFINITIONS
// Each of these typedefs mirrors as closely as possible the
// declarations found in malloc.h and stdlib.h. Use them to
// point to your own versions of each routine.
typedef void * SysCallocFn  ( size_t num, size_t size );
typedef void * SysMallocFn  ( size_t size );
typedef void * SysReallocFn ( void *memblock, size_t size );
typedef void   SysFreeFn    ( void *memblock );
typedef int    SysHeapChkFn ( void );
typedef void   SysHeapDumpFn( void );

#ifndef BOUNDS_CHECKER

	extern void * sysMalloc    ( size_t size );

	#if 0
		extern void *hackMalloc( char *fileLine, int bytes );
		#define sysMalloc(n)	hackMalloc(__FILE__##"__LINE__",n)
	#endif

	// DECLARATIONS
	// These are the memory management routines you should call.
	extern void * sysCalloc    ( size_t num, size_t size );
	extern void * sysRealloc   ( void *memblock, size_t size );
 	extern void   sysFree      ( void *memblock );
	extern int    sysHeapChk   ( void );
	extern void   sysHeapDump  ( void );

	#define       sysAlloca(size)   alloca(size)
	// Why isn't sysAlloca just another function call?  Because it allocates
	// memory on the stack, and under certain compilation models, it MUST
	// be called directly to work right - in other words, calling it through
	// an intermediate function will fail.

	extern char * sysStrDup    ( const char *src );
	// This overloads strdup() because that routine called malloc().

#else
	#include "malloc.h"

	// DECLARATIONS
	// These are the memory management routines you should call.
	#define sysCalloc     calloc
	#define sysMalloc     malloc
	#define sysRealloc 	  realloc
	#define sysFree       free
	#define sysHeapChk   
	#define sysHeapDump  
	#define sysAlloca(n)  alloca(n)
	#define sysStrDup     strdup

#endif


// INITIALIZER
// Initialize the mem management routines here.  If you do not call
// this routine, the VC memory routines will be called by default.
// Pass in NULL for any routine you don't want to overload.
extern void sysMallocInit( 
		SysCallocFn   *calloc, 
		SysMallocFn   *malloc, 
		SysReallocFn  *realloc,
		SysFreeFn     *free,
		SysHeapChkFn  *heapChk,
		SysHeapDumpFn *heapDump
	);

// STOLEN DEFINITION
// Stolen from malloc.h to define alloca
#ifndef _POSIX_
	#if     !__STDC__
		#ifdef __cplusplus
		extern "C" {
		#endif
			void *          __cdecl _alloca(size_t);
		#ifdef __cplusplus
		}
		#endif
		/* Non-ANSI names for compatibility */
		#define alloca  _alloca
	#endif  /* __STDC__*/

	#if defined(_M_MRX000) || defined(_M_PPC) || defined(_M_ALPHA)
		#pragma intrinsic(_alloca)
	#endif
#endif  /* _POSIX_ */


#endif


/*
Example Call to Allocate Memory:

char *s = (char *)sysMalloc( 1024 );

Example of Overloading sysMalloc:

extern "C" void *sysMalloc( size_t size );
sysMalloc = sysMalloc;

void *sysMalloc( size_t size )
{
	printf( "You just allocated %d bytes!\n", size );
	#ifdef DEBUG
		return __sysMallocDebug( size );
	#else
		return __sysMalloc( size );
	#endif
}

*/
