#ifdef WIN32
	#include "windows.h"	// All just for timeGetTime()!!!!
	#include "mmsystem.h"
	#include "time.h"

	unsigned int milsAtStart = 0;
	unsigned int secsAtStart = 0;

#else
	#include "sys/time.h"
	#include "unistd.h"
#endif

#include "fulltime.h"
#include "math.h"
#include "assert.h"
//#include "trace.h"

#define JAN_1_97_IN_UTC ( ( 27*365 + (27/4) )*24*60*60 )


// Global Variables
//------------------------------------------------------------------------

int globalFrameNumber = 0;

int serverDiffMils = 0;

FullTime frameTime = 0.0;
FullTime uncorrectedFrameTime = 0.0;
FullTime lastFrameElapsedTime = 0.0;
FullTime firstTime = 0.0;

int timeStopped = 0;
FullTime whenStopped = 0.0;
FullTime timeStoppedAccumulator = 0.0;

double thisFPS = 0.0;
double avgFPS = 0.0;
double avgSPF = 0.0;


// Interface Functions
//------------------------------------------------------------------------

void getSecsAndMils( unsigned int &secs, unsigned int &mils, int correct ) {
	// This function hides the OS dependence problems
	int _serverDiffMils = correct ? serverDiffMils : 0;
	#ifdef WIN32
		#if 0
			SYSTEMTIME t;
			GetSystemTime( &t );
			secs = t.wHour*3600 + t.wMinute*60 + t.wSecond;
			mils = t.wMilliseconds;
		#else
			__int64 _mils = timeGetTime() - milsAtStart;
			_mils -= _serverDiffMils;
			_mils += (__int64)secsAtStart * (__int64)1000;
			secs = (unsigned int)_mils / 1000;
			mils = (unsigned int)_mils % 1000;
//			mils = timeGetTime() - milsAtStart - _serverDiffMils;
//			secs = secsAtStart + (mils / 1000);
//			mils %= 1000;
		#endif
	#else
		// *** Not correcting under UNIX
		struct timeval tv;
		struct timezone tz;
		gettimeofday( &tv, &tz );
		mils = tv.tv_usec/1000;
		secs = tv.tv_sec - JAN_1_97_IN_UTC;
	#endif
}	


FullTime getRealTime() {
	assert( serverDiffMils == 0.0 || timeStoppedAccumulator == 0.0 );
		// If the time synchronizer is active, you can not
		// use the time stopper.  Be sure to call resetStoppedTime()
		// before using the time synchronizer
	
	if( timeStopped ) {
		return whenStopped;
	}
	unsigned int secs, mils;
	getSecsAndMils( secs, mils, 1 );
	return FullTime( (double)secs + (double)mils/1000.0 ) - timeStoppedAccumulator;
}

FullTime getUncorrectedTime() {
	unsigned int secs, mils;
	getSecsAndMils( secs, mils, 0 );
	return FullTime( (double)secs + (double)mils/1000.0 );
}

int isTimeStopped() {
	return timeStopped > 0;
}

void stopTime() {
	if( timeStopped == 0 ) {
		whenStopped = getRealTime();
	}
	++timeStopped;
}

void resumeTime()
{
	if( timeStopped )
	{
		--timeStopped;
		if( !timeStopped )
		{
			timeStoppedAccumulator.time += getRealTime() - whenStopped;
		}
	}
}

void resetStoppedTime()
{
	timeStoppedAccumulator = 0.0;
}

void setFrameTime()
{
	lastFrameElapsedTime = frameTime;
		// A temporary storage place

	// Set the frameTime and uncorrectedFrameTime
 	frameTime = getRealTime();
 	uncorrectedFrameTime = getUncorrectedTime();

	if( firstTime == 0.0 )
	{
		firstTime = frameTime;
	}

	lastFrameElapsedTime = frameTime - lastFrameElapsedTime;

	// Stick uncorrected time into the rolling queue for averaging purposes
	static FullTime frameTimes[FRAMES_TO_AVG];
	frameTimes[globalFrameNumber%FRAMES_TO_AVG] = uncorrectedFrameTime;

	// Set the current fps as long as we have one sample
	if( globalFrameNumber > 0 ) {
		// For some reason, two frames were reporting identical times,
		// leading to a divide by zero.  We check for this now.
		double frameDelta = (
			uncorrectedFrameTime - 
			frameTimes[ (globalFrameNumber+FRAMES_TO_AVG-1) % FRAMES_TO_AVG ]
		);

		// NOTE: -1.0 is a magic value used only when we would divide by zero
		thisFPS = -1.0;
		if (frameDelta > 0.0) {
			thisFPS = 1.0 / frameDelta;
		}
	}

	// Set the average as long as we have enough samples
	if( globalFrameNumber > FRAMES_TO_AVG ) {
		double elapsed = uncorrectedFrameTime -
			frameTimes[ (globalFrameNumber+1) % FRAMES_TO_AVG ]
		;

		// NOTE: -1.0 is a magic value used only when we would divide by zero
		avgFPS = -1.0;
		if (elapsed > 0.0) {
			avgFPS = (double)FRAMES_TO_AVG / elapsed;
		}
		avgSPF = 1.0/avgFPS;
	}

	globalFrameNumber++;
}

void initFullTime()
{ 
	// The timeGetTime has the best accuracy, but it doesn't give the
	// time in UTC, so we get that at the begining and the delta off of that

 	#ifdef WIN32
		HANDLE process = GetCurrentProcess();
		DWORD oldPriority = GetPriorityClass( process );
		SetPriorityClass( process, HIGH_PRIORITY_CLASS );

		SYSTEMTIME t;
		GetSystemTime( &t );
		int lastSec = t.wSecond;
		while( 1 ) {
			GetSystemTime( &t );
			if( t.wSecond != lastSec ) {
				milsAtStart = timeGetTime();
				secsAtStart = time(NULL) - JAN_1_97_IN_UTC;
				break;
			}
		}
		
		setFrameTime();

		SetPriorityClass( process, oldPriority );
	#endif
} 

