#include "TimeStr.h"
#include "UniversalTime.h"
#include "FileTime.h"
#include "time.h"
//#include "StrExtra.h"
#include "stdlib.h"
#include "string.h"
#include "stdio.h"
#include "minmax.h"
#include "SysAssert.h"


char *monthAbbrev[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
char *dayAbbrev[7]    = { "Mon", "Tue", "Wed", "Thu", "Fri" };

char *getMonth( int n )
{
	sysAssert( n>=0 && n<12 );
	return monthAbbrev[n];
}

int test( UniversalTime *u )
{
	int ok = 1;
	ok = ok && ( u->year >= 0 && u->year < 3000 );
	ok = ok && ( u->month >= 1 && u->month <= 12 );
	ok = ok && ( u->day >= 1 && u->day <= 31 );
	ok = ok && ( u->hour >= 0 && u->hour <= 23 );
	ok = ok && ( u->minute >= 0 && u->minute <= 59 );
	ok = ok && ( u->second >= 0 && u->second <= 59 );
	ok = ok && ( u->millisecond >= 0 && u->millisecond <= 999 );
	return ok;
}

void zero( UniversalTime *u )
{
	u->year = 0;
	u->month = 1;
	u->day = 1;
	u->hour = 0;
	u->minute = 0;
	u->second = 0;
	u->millisecond = 0;
}


TimeStr::TimeStr()
{
	str[0] = '\0';
	u = new UniversalTime;
	FileTime f;
	fileTimeGetUTC( &f );
	fileTimeToUniversalTime( &f, u );
	sysAssert( test( u ) );
}

void TimeStr::set( TimeStr &t )
{
	sysAssert( u );
	*u = *t.u;
	strcpy( str, t.str );
}


TimeStr &TimeStr::getLocal()
{
	FileTime f;
	str[0] = '\0';
	fileTimeGetLocal( &f );
	fileTimeToUniversalTime( &f, u );
	sysAssert( test( u ) );
	return *this;
}


TimeStr::~TimeStr()
{
	delete u;
}												  

TimeStr::TimeStr( UniversalTime *_u )
{
	u = new UniversalTime;
	*u = *_u;
	if( !test( u ) )
		zero( u );
}

TimeStr::TimeStr( FileTime *fTime )
{
	u = new UniversalTime;
	memset( u, 0, sizeof( *u ) );
	fileTimeToUniversalTime( fTime, u );
	if( !test( u ) )
		zero( u );
}

TimeStr::TimeStr( struct tm *tmTime )
{
	u = new UniversalTime;
	memset( u, 0, sizeof( *u ) );
    u->year        = tmTime->tm_year + 1900;
    u->month       = tmTime->tm_mon;
    u->dayOfWeek   = tmTime->tm_wday;
    u->day         = tmTime->tm_mday;
    u->hour        = tmTime->tm_hour;
    u->minute      = tmTime->tm_min;
    u->second      = tmTime->tm_sec;
	u->millisecond = 0;
	if( !test( u ) )
		zero( u );
}

char *timeStrcpyUntil( char *dest, char *src, char *stoppers )
{
	while( *src && !strchr( stoppers, *src ) )
	{
		*dest = *src;
		++src;
		++dest;
	}
	*dest = '\0';

	return src;
}

int strInt( char* &s )
{
	char temp[20];
	s = timeStrcpyUntil( temp, s, " :-,." );
	if( *s ) ++s;
	return atoi( temp );
}

TimeStr::TimeStr( char *httpTime )
{
	char dow[10];
	char month[10];
	
	u = new UniversalTime;
	memset( u, 0, sizeof(*u) );

	if( !httpTime )
		return;

	char *s = httpTime;
	s = timeStrcpyUntil( dow, s, " " );
	if( *s ) ++s;
	u->day = strInt( s );
	s = timeStrcpyUntil( month, s, " :-" );
	if( *s ) ++s;
	u->year = strInt( s );
	if( u->year < 50 )
		u->year += 2000;
	if( u->year < 100 )
		u->year += 1900;
	u->hour = strInt( s );
	u->minute = strInt( s );
	u->second = strInt( s );

	for( u->month = 1 ; u->month <= 12 ; ++u->month )
	{
		if( !stricmp( getMonth(u->month-1), month ) )
		{
			break;
		}
	}
	if( !test( u ) )
		zero( u );
}

void TimeStr::setElapsed( char *elapsedTime )
{
	memset( u, 0, sizeof(*u) );

	if( !elapsedTime )
		return;

	u->month = 1;
	u->day = 1;

	char *s = elapsedTime;
	u->hour = strInt( s );
	u->minute = strInt( s );
	u->second = strInt( s );
	u->millisecond = strInt( s );
	if( !test( u ) )
		zero( u );
}



char *TimeStr::get()
{
// Clear time format: 15-Nov-94 15:45:12

	if( str[0] == '\0' )
	{
		sysAssert( test( u ) );
		sprintf( str, "%02d-%s-%d %02d:%02d:%02d",
			u->day,
			getMonth(u->month-1),
			u->year,
			u->hour,
			u->minute,
			u->second
		);
	}
	return str;
}

char *TimeStr::getSortable()
{
// Sortable time format: 1999-11-15-15-45-12

	sysAssert( test( u ) );
	sprintf( str, "%04d-%02d-%02d-%02d-%02d-%02d",
		u->year,
		u->month,
		u->day,
		u->hour,
		u->minute,
		u->second
	);
	return str;
}

char *TimeStr::getElapsed()
{
// Elapsed time 15:45:12.321
	sysAssert( test( u ) );
	sprintf( str, "%02d:%02d:%02d.%03d",
		u->hour,
		u->minute,
		u->second,
		u->millisecond
	);
	return str;
}

char *TimeStr::getHttp()
{
// HTTP time format: Tue, 15 Nov 1994 12:45:26 GMT
	sysAssert( test( u ) );
	sprintf( str, "%s, %02d %s %d %02d:%02d:%02d GMT",
		dayAbbrev[min(6,max(0,u->dayOfWeek))],
		u->day,
		getMonth(u->month-1),
		u->year,
		u->hour,
		u->minute,
		u->second
	);
	return str;
}

char *TimeStr::getSqlTime()
{
// Sql time is 'YYYY-MM-DD HH:MM:SS' 
	sysAssert( test( u ) );
	sprintf( str, "%04d-%02d-%02d %02d:%02d:%02d",
		u->year,
		u->month,
		u->day,
		u->hour,
		u->minute,
		u->second
	);
	return str;
}

char *TimeStr::getAsNumber()			// YYYYMMDDHHMMSS
{
	sysAssert( test( u ) );
	sprintf( str, "%04d%02d%02d%02d%02d%02d",
		u->year,
		u->month,
		u->day,
		u->hour,
		u->minute,
		u->second
	);
	return str;
}


char *TimeStr::getDate()
{
// HTTP time format: 15-Nov-1994
	sysAssert( test( u ) );
	sprintf( str, "%02d-%s-%d",
		u->day,
		getMonth(u->month-1),
		u->year
	);
	return str;
}

char *TimeStr::getSortableDate()
{
// HTTP time format: YYYY-MM-DD
	sysAssert( test( u ) );
	sprintf( str, "%04d-%02d-%02d",
		u->year,
		u->month,
		u->day
	);
	return str;
}


char *TimeStr::getTime()
{
// HTTP time format: 14:22
	sysAssert( test( u ) );
	sprintf( str, "%02d:%02d",
		u->hour,
		u->minute
	);
	return str;
}


char *TimeStr::getHttpShort()
{
// HTTP short time format: 15 Nov 1994 12:45:26
	sysAssert( test( u ) );
	sprintf( str, "%02d %s %04d %02d:%02d:%02d",
		u->day,
		getMonth(u->month-1),
		u->year,
		u->hour,
		u->minute,
		u->second
	);
	return str;
}

int TimeStr::getElapsedInt()
{
	return u->hour*60*60*1000 +
		u->minute*60*1000 +
		u->second*1000 +
		u->millisecond;
}

TimeStr &TimeStr::operator = (TimeStr &t)
{
	set( t );
	return *this;
}

