#include "UniversalTime.h"
#include "FileTime.h"
#include "windows.h"
#include "SysAssert.h"


// For SYSTEMTIME, 
// wMonth     = The current month; January is 1.
// wDayOfWeek = The current day of the week; Sunday is 0, Monday is 1, and so on.



void systemTimeToUniversalTime( SYSTEMTIME *winSystemTime, UniversalTime *ut )
{
	ut->year = winSystemTime->wYear;
	ut->month = winSystemTime->wMonth;
	ut->dayOfWeek = winSystemTime->wDayOfWeek;
	ut->day = winSystemTime->wDay;
	ut->hour = winSystemTime->wHour;
	ut->minute = winSystemTime->wMinute;
	ut->second = winSystemTime->wSecond;
	ut->millisecond = winSystemTime->wMilliseconds;

}

void universalTimeToSystemTime( UniversalTime *ut, SYSTEMTIME *winSystemTime )
{
	winSystemTime->wYear = ut->year;
	winSystemTime->wMonth = ut->month;
	winSystemTime->wDayOfWeek = ut->dayOfWeek;
	winSystemTime->wDay = ut->day;
	winSystemTime->wHour = ut->hour;
	winSystemTime->wMinute = ut->minute;
	winSystemTime->wSecond = ut->second;
	winSystemTime->wMilliseconds = ut->millisecond;
}


void fileTimeToUniversalTime( FileTime *ft, UniversalTime *ut )
{
	FILETIME   winFileTime;
	SYSTEMTIME winSystemTime;

	winFileTime.dwHighDateTime = ft->hi;
	winFileTime.dwLowDateTime =  ft->lo;

	FileTimeToSystemTime( &winFileTime, &winSystemTime );
	systemTimeToUniversalTime( &winSystemTime, ut );
}

void universalTimeToFileTime( UniversalTime *ut, FileTime *ft )
{
	FILETIME   winFileTime;
	SYSTEMTIME winSystemTime;

	winFileTime.dwHighDateTime = 0;
	winFileTime.dwLowDateTime  = 0;

	universalTimeToSystemTime( ut, &winSystemTime );
	SystemTimeToFileTime( &winSystemTime, &winFileTime );

	ft->lo = winFileTime.dwLowDateTime;
	ft->hi = winFileTime.dwHighDateTime;

	UniversalTime check;
	fileTimeToUniversalTime( ft, &check );

	if( ut->year > 1600 )
		sysAssert( universalTimeEquals( *ut, check ) );

}

int universalTimeEquals(  UniversalTime remote, UniversalTime local, int testMilli )
{
	if( remote.year   == local.year &&
		remote.month  == local.month &&
		remote.day    == local.day &&
		remote.hour   == local.hour &&
		remote.minute == local.minute &&
		remote.second == local.second &&
		(!testMilli || (remote.millisecond == local.millisecond) )
		)
		return 1;
	return 0;
}

int universalTimeLaterThan( UniversalTime isThisLater, UniversalTime thanThis )
{
	if( isThisLater.year < thanThis.year )
		return 0;
	if( isThisLater.year == thanThis.year )
	{
		if( isThisLater.month < thanThis.month )
			return 0;
		if( isThisLater.month == thanThis.month )
		{
			if( isThisLater.day < thanThis.day )
				return 0;
			if( isThisLater.day == thanThis.day )
			{
				if( isThisLater.hour < thanThis.hour )
					return 0;
				if( isThisLater.hour == thanThis.hour )
				{
					if( isThisLater.minute < thanThis.minute )
						return 0;
					if( isThisLater.minute == thanThis.minute )
					{
						if( isThisLater.second < thanThis.second )
							return 0;
						if( isThisLater.second == thanThis.second )
						{
							if( isThisLater.millisecond <= thanThis.millisecond )
								return 0;
						}
					}
				}
			}
		}
	}
	return 1;
}